from office365.runtime.client_result import ClientResult
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


def create_download_content_query(file_item, format_name=None):
    """
    :type file_item: office365.onedrive.driveItem.DriveItem
    :type format_name: str or None
    """
    result = ClientResult(file_item.context)
    action_name = "content"
    if format_name is not None:
        action_name = action_name + r"?format={0}".format(format_name)
    qry = ServiceOperationQuery(file_item, action_name, None, None, None, result)

    def _construct_query(request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions
        """
        request.method = HttpMethod.Get
    file_item.context.before_execute(_construct_query)
    return qry




from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity


class ListItemVersion(BaseEntity):
    """Represents a version of a list item."""

    @property
    def version_id(self):
        """Gets the ID of the version."""
        return int(self.properties.get("VersionId", -1))

    @property
    def version_label(self):
        """Gets the version number of the item version."""
        return self.properties.get("VersionLabel")

    @property
    def is_current_version(self):
        """Gets a value that specifies whether the file version is the current version.

        :rtype: bool
        """
        return self.properties.get("IsCurrentVersion", None)

    @property
    def created(self):
        """Gets the creation date and time for the item version."""
        return self.properties.get("Created", None)

    @property
    def created_by(self):
        """Gets the user that created the item version."""
        from office365.sharepoint.principal.user import User
        return self.properties.get("CreatedBy", User(self.context, ResourcePath("CreatedBy", self.resource_path)))

    @property
    def fields(self):
        """Gets the collection of fields that are used in the list that contains the item version."""
        from office365.sharepoint.fields.field_collection import FieldCollection
        return self.properties.get("Fields", FieldCollection(self.context, ResourcePath("Fields", self.resource_path)))

    @property
    def file_version(self):
        from office365.sharepoint.files.file_version import FileVersion
        return self.properties.get("FileVersion",
                                   FileVersion(self.context, ResourcePath("FileVersion", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "CreatedBy": self.created_by,
                "FileVersion": self.file_version
            }
            default_value = property_mapping.get(name, None)
        return super(ListItemVersion, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        if self._resource_path is None:
            if name == "VersionId":
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)
        return super(ListItemVersion, self).set_property(name, value, persist_changes)





from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.planner.plans.plan import PlannerPlan
from office365.planner.tasks.task import PlannerTask
from office365.runtime.resource_path import ResourcePath


class PlannerUser(Entity):
    """The plannerUser resource provide access to Planner resources for a user.
    It doesn't contain any usable properties."""

    @property
    def plans(self):
        """Read-only. Nullable. Returns the plannerTasks assigned to the user.

        :rtype: EntityCollection
        """
        return self.get_property('plans',
                                 EntityCollection(self.context, PlannerPlan,
                                                  ResourcePath("plans", self.resource_path)))

    @property
    def tasks(self):
        """Read-only. Nullable. Returns the plannerTasks assigned to the user.

        :rtype: EntityCollection
        """
        return self.get_property('tasks',
                                 EntityCollection(self.context, PlannerTask,
                                                  ResourcePath("tasks", self.resource_path)))



from office365.entity import Entity


class TeamsAsyncOperation(Entity):
    """
    A Microsoft Teams async operation is an operation that transcends the lifetime of a single API request.
    These operations are long-running or too expensive to complete within the timeframe of their originating request.

    When an async operation is initiated, the method returns a 202 Accepted response code.
    The response will also contain a Location header, which contains the location of the teamsAsyncOperation.
    Periodically check the status of the operation by making a GET request to this location; wait >30 seconds
    between checks. When the request completes successfully, the status will be "succeeded" and
    the targetResourceLocation will point to the created/modified resource.

    """
    pass



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.contenttypes.field_link import FieldLink


class FieldLinkCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(FieldLinkCollection, self).__init__(context, FieldLink, resource_path)



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.principal.user import User


class RecycleBinItem(BaseEntity):
    """Represents a Recycle Bin item in the Recycle Bin of a site or a site collection."""

    def restore(self):
        """Restores the Recycle Bin item to its original location."""
        qry = ServiceOperationQuery(self, "Restore")
        self.context.add_query(qry)
        return self

    def move_to_second_stage(self):
        qry = ServiceOperationQuery(self, "MoveToSecondStage")
        self.context.add_query(qry)
        return self

    @property
    def id(self):
        """Gets a value that specifies the identifier of the Recycle Bin item."""
        return self.properties.get('Id', None)

    @property
    def size(self):
        """Gets a value that specifies the size of the Recycle Bin item in bytes.

        :rtype: long or None
        """
        return self.properties.get('Size', None)

    @property
    def author(self):
        """
        Gets a value that specifies the user who created the Recycle Bin item.

        :rtype: office365.sharepoint.directory.user.User or None
        """
        return self.properties.get("Author", User(self.context, ResourcePath("Author", self.resource_path)))

    @property
    def deleted_by(self):
        """Gets a value that specifies the user who deleted the Recycle Bin item."""
        return self.properties.get('DeletedBy', User(self.context, ResourcePath("DeletedBy", self.resource_path)))

    @property
    def deleted_date(self):
        """Gets a value that specifies when the Recycle Bin item was moved to the Recycle Bin."""
        return self.properties.get('DeletedDate', None)

    def set_property(self, name, value, persist_changes=True):
        super(RecycleBinItem, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if self._resource_path is None:
            if name == "Id" and self._parent_collection is not None:
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)



from office365.directory.directory_object_collection import DirectoryObjectCollection
from office365.entity import Entity
from office365.runtime.resource_path import ResourcePath


class Directory(Entity):
    """Represents a deleted item in the directory. When an item is deleted, it is added to the deleted items
    "container". Deleted items will remain available to restore for up to 30 days. After 30 days, the items are
    permanently deleted. """

    def deleted_items(self, entity_type=None):
        """Recently deleted items. Read-only. Nullable."""
        if entity_type:
            return DirectoryObjectCollection(self.context,
                                             ResourcePath(entity_type,
                                                          ResourcePath("deletedItems", self.resource_path)))
        else:
            return self.properties.get('deletedItems',
                                       DirectoryObjectCollection(self.context,
                                                                 ResourcePath("deletedItems", self.resource_path)))

    @property
    def deleted_groups(self):
        """Recently deleted groups"""
        return self.deleted_items("microsoft.graph.group")

    @property
    def deleted_users(self):
        """Recently deleted users"""
        return self.deleted_items("microsoft.graph.user")

    @property
    def deleted_applications(self):
        """Recently deleted applications"""
        return self.deleted_items("microsoft.graph.application")



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.files.file import File


def create_file_query(files, file_creation_information):
    """
    :type file_creation_information: office365.sharepoint.files.file_creation_information.FileCreationInformation
    :type files: office365.sharepoint.files.file_collection.FileCollection
    """
    return_file = File(files.context)
    qry = ServiceOperationQuery(files, "add", file_creation_information.to_json(),
                                file_creation_information.content, None,
                                return_file)
    files.add_child(return_file)
    return qry



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity


class ClientWebPart(BaseEntity):

    def render(self, properties=None):
        """
        :param dict properties:
        """
        return_type = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "Render", None, properties, None, return_type)
        self.context.add_query(qry)
        return return_type



from office365.communications.callrecords.call_record import CallRecord
from office365.communications.calls.call import Call
from office365.communications.onlinemeetings.online_meeting_collection import OnlineMeetingCollection
from office365.communications.presences.presence import Presence
from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class CloudCommunications(Entity):

    def get_presences_by_user_id(self, ids):
        """
        Get the presence information for multiple users.

        :type ids: list[str]
        """
        return_type = EntityCollection(self.context, Presence, ResourcePath("presences", self.resource_path))
        qry = ServiceOperationQuery(self, "getPresencesByUserId", None, {"ids": ids}, None, return_type)
        self.context.add_query(qry)
        return return_type

    @property
    def calls(self):
        """"
        :rtype: EntityCollection
        """
        return self.get_property('calls',
                                 EntityCollection(self.context, Call, ResourcePath("calls", self.resource_path)))

    @property
    def call_records(self):
        """"
        :rtype: EntityCollection
        """
        return self.get_property('callRecords',
                                 EntityCollection(self.context, CallRecord,
                                                  ResourcePath("callRecords", self.resource_path)))

    @property
    def online_meetings(self):
        """"
        :rtype: OnlineMeetingCollection
        """
        return self.get_property('onlineMeetings',
                                 OnlineMeetingCollection(self.context,
                                                         ResourcePath("onlineMeetings", self.resource_path)))

    @property
    def presences(self):
        """"
        :rtype: EntityCollection
        """
        return self.get_property('presences',
                                 EntityCollection(self.context, Presence,
                                                  ResourcePath("presences", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "callRecords": self.call_records,
                "onlineMeetings": self.online_meetings
            }
            default_value = property_mapping.get(name, None)
        return super(CloudCommunications, self).get_property(name, default_value)



from office365.runtime.client_object import ClientObject


class ClientSidePage(ClientObject):
    """Represents the data and methods associated with client side "modern" pages"""

    @staticmethod
    def from_file(file):
        return None



from office365.base_item import BaseItem
from office365.directory.subscriptions.subscription import Subscription
from office365.entity_collection import EntityCollection
from office365.onedrive.columns.column_definition import ColumnDefinition
from office365.onedrive.contenttypes.content_type import ContentType
from office365.onedrive.listitems.list_item import ListItem
from office365.runtime.client_value import ClientValue
from office365.runtime.resource_path import ResourcePath


class ListInfo(ClientValue):

    def __init__(self, template=None, content_types_enabled=False, hidden=False):
        super(ListInfo, self).__init__()
        self.template = template
        self.contentTypesEnabled = content_types_enabled
        self.hidden = hidden


class List(BaseItem):
    """The list resource represents a list in a site. This resource contains the top level properties of the list,
    including template and field definitions. """

    @property
    def list(self):
        """Provides additional details about the list."""
        return self.properties.get('list', ListInfo())

    @property
    def sharepoint_ids(self):
        """Returns identifiers useful for SharePoint REST compatibility."""
        return self.properties.get('sharepointIds', None)

    @property
    def drive(self):
        """Only present on document libraries. Allows access to the list as a drive resource with driveItems."""
        from office365.onedrive.drives.drive import Drive
        return self.get_property('drive',
                                 Drive(self.context, ResourcePath("drive", self.resource_path)))

    @property
    def columns(self):
        """The collection of columns under this site."""
        return self.get_property('columns',
                                 EntityCollection(self.context, ColumnDefinition,
                                                  ResourcePath("columns", self.resource_path)))

    @property
    def content_types(self):
        """The collection of content types under this site."""
        return self.get_property('contentTypes',
                                 EntityCollection(self.context, ContentType,
                                                  ResourcePath("contentTypes", self.resource_path)))

    @property
    def items(self):
        """All items contained in the list."""
        return self.get_property('items',
                                 EntityCollection(self.context, ListItem, ResourcePath("items", self.resource_path)))

    @property
    def subscriptions(self):
        """The set of subscriptions on the list."""
        return self.get_property('subscriptions',
                                 EntityCollection(self.context, Subscription,
                                                  ResourcePath("subscriptions", self.resource_path)))



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity_collection import BaseEntityCollection


class ViewFieldCollection(BaseEntityCollection):
    """Represents a collection of Field resources."""
    def __init__(self, context, resource_path=None):
        super(ViewFieldCollection, self).__init__(context, str, resource_path)

    @property
    def schema_xml(self):
        """Gets Schema Xml.

        :rtype: str or None
        """
        return self.properties.get('SchemaXml', None)

    @property
    def items(self):
        """Gets items.

        :rtype: list[str] or None
        """
        return self.properties.get('Items', None)

    def set_property(self, name, value, persist_changes=False):
        self._properties[name] = value
        if name == "Items":
            self._data = list(value.values())
        return self

    def add_view_field(self, field_name):
        """
        Adds the field with the specified field internal name or display name to the collection.

        :param str field_name:
        :return:
        """
        qry = ServiceOperationQuery(self, "AddViewField", [field_name])
        self.context.add_query(qry)
        return self

    def move_view_field_to(self, name, index):
        """
        Moves the field with the specified field internal name to the specified position in the collection

        :param str name: Specifies the field internal name.
        :param int index: Specifies the new position for the field (2). The first position is 0.

        """
        params = {"field": name, "index": index}
        qry = ServiceOperationQuery(self, "MoveViewFieldTo", None, params)
        self.context.add_query(qry)
        return self

    def remove_all_view_fields(self):
        """
        Removes all the fields from the collection.

        """
        qry = ServiceOperationQuery(self, "RemoveAllViewFields")
        self.context.add_query(qry)
        return self

    def remove_view_field(self, field_name):
        """
        Removes the field with the specified field internal name or display name from the collection.

        :param str field_name:
        """
        qry = ServiceOperationQuery(self, "RemoveViewField", [field_name])
        self.context.add_query(qry)
        return self



from office365.onedrive.versions.base_item_version import BaseItemVersion
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class DriveItemVersion(BaseItemVersion):
    """The DriveItemVersion resource represents a specific version of a DriveItem."""

    def restore_version(self):
        qry = ServiceOperationQuery(self, "restoreVersion")
        self.context.add_query(qry)
        return self

    @property
    def content(self):
        """
        The content stream for this version of the item.

        :rtype: str or bytes
        """
        return self.properties.get('content', None)

    @property
    def size(self):
        """
        Indicates the size of the content stream for this version of the item.

        :rtype: int
        """
        return self.properties.get('size', None)



from office365.entity_collection import EntityCollection
from office365.outlook.calendar.calendar import Calendar
from office365.entity import Entity
from office365.runtime.resource_path import ResourcePath


class CalendarGroup(Entity):
    """
    A group of user calendars.
    """

    @property
    def calendars(self):
        """The calendars in the calendar group. Navigation property. Read-only. Nullable."""
        return self.properties.get('calendars',
                                   EntityCollection(self.context, Calendar,
                                                    ResourcePath("calendars", self.resource_path)))



from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.resource_path import ResourcePath


class ContactFolder(Entity):
    """A folder that contains contacts."""

    @property
    def contacts(self):
        """The contacts in the folder. Navigation property. Read-only. Nullable."""
        from office365.outlook.contacts.contact import Contact
        return self.properties.get('contacts',
                                   EntityCollection(self.context, Contact,
                                                    ResourcePath("contacts", self.resource_path)))

    @property
    def child_folders(self):
        """The collection of child folders in the folder. Navigation property. Read-only. Nullable."""
        return self.properties.get('childFolders',
                                   EntityCollection(self.context, ContactFolder,
                                                    ResourcePath("childFolders", self.resource_path)))



import json

from office365.runtime.client_value import ClientValue
from office365.runtime.compat import is_string_type


class ODataPathParser(object):
    @staticmethod
    def parse_path_string(string):
        pass

    @staticmethod
    def from_method(method_name, method_parameters=None):
        """

        :type method_parameters: list or dict or ClientValue
        :type method_name: str
        """
        url = ""
        if method_name:
            url = method_name

        if isinstance(method_parameters, ClientValue):
            param_value = method_parameters.to_json()
            url += "(@v)?@v={0}".format(json.dumps(param_value))
        elif method_parameters is not None:
            url += "("
            if isinstance(method_parameters, dict):
                url += ','.join(['%s=%s' % (key, ODataPathParser.encode_method_value(value)) for (key, value) in
                                 method_parameters.items() if value is not None])
            else:
                url += ','.join(['%s' % (ODataPathParser.encode_method_value(value)) for (i, value) in
                                 enumerate(method_parameters) if value is not None])
            url += ")"
        return url

    @staticmethod
    def encode_method_value(value):
        if is_string_type(value):
            value = value.replace("'", "''")

            # Same replacements as SQL Server
            # https://web.archive.org/web/20150101222238/http://msdn.microsoft.com/en-us/library/aa226544(SQL.80).aspx
            # https://stackoverflow.com/questions/4229054/how-are-special-characters-handled-in-an-odata-query#answer-45883747
            value = value.replace('%', '%25')
            value = value.replace('+', '%2B')
            value = value.replace('/', '%2F')
            value = value.replace('?', '%3F')
            value = value.replace('#', '%23')
            value = value.replace('&', '%26')

            value = "'{0}'".format(value)
        elif isinstance(value, bool):
            value = str(value).lower()
        return value



from office365.runtime.client_value import ClientValue


class SharingLinkAccessRequest(ClientValue):

    def __init__(self):
        super(SharingLinkAccessRequest, self).__init__()
        self.ensureAccess = None
        self.password = None



from office365.runtime.client_value import ClientValue


class XmlSchemaFieldCreationInformation(ClientValue):

    def __init__(self, schemaXml=None, options=None):
        """
        :type schemaXml: str
        :type options: int or None
        """
        super(XmlSchemaFieldCreationInformation, self).__init__()
        self.SchemaXml = schemaXml
        self.Options = options

    @property
    def entity_type_name(self):
        return "SP.XmlSchemaFieldCreationInformation"



from office365.runtime.client_value import ClientValue


class Folder(ClientValue):
    """
    The Folder resource groups folder-related data on an item into a single structure.
    DriveItems with a non-null folder facet are containers for other DriveItems.
    """

    def __init__(self, childCount=None, view=None):
        """

        :param int childCount:
        :param office365.onedrive.folderView.FolderView view:
        """
        super(Folder, self).__init__()
        self.childCount = childCount
        self.view = view



from office365.runtime.client_runtime_context import ClientRuntimeContext


class ExcelService(ClientRuntimeContext):

    def __init__(self, context):
        """
        Excel Services REST API client
        https://docs.microsoft.com/en-us/sharepoint/dev/general-development/excel-services-rest-api
        """
        super(ExcelService, self).__init__()

    def authenticate_request(self, request):
        pass

    def service_root_url(self):
        return "{0}/_vti_bin/ExcelRest.aspx"

    def pending_request(self):
        pass

    def get_workbook(self, list_name, file_name):
        return self



from office365.runtime.client_value import ClientValue


class TeamsTabConfiguration(ClientValue):
    """The settings that determine the content of a tab.
    When a tab is interactively configured, this information is set by the tab provider application.
    In addition to the properties below, some tab provider applications specify additional custom properties.
    """
    pass



from office365.outlook.calendar.calendar_permission import CalendarPermission
from office365.outlook.calendar.dateTimeTimeZone import DateTimeTimeZone
from office365.outlook.calendar.email_address import EmailAddress
from office365.outlook.calendar.event import Event
from office365.outlook.calendar.schedule_information import ScheduleInformation
from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class Calendar(Entity):
    """
    A calendar which is a container for events. It can be a calendar for a user, or the default calendar
        of a Microsoft 365 group.
    """

    def get_schedule(self, schedules, start_time=None, end_time=None, availability_view_interval=30):
        """
        Get the free/busy availability information for a collection of users, distributions lists, or resources
        (rooms or equipment) for a specified time period.

        :param datetime.datetime end_time: The date, time, and time zone that the period ends.
        :param int availability_view_interval: Represents the duration of a time slot in an availabilityView
             in the response. The default is 30 minutes, minimum is 5, maximum is 1440. Optional.
        :param datetime.datetime start_time: The date, time, and time zone that the period starts.
        :param list[str] schedules: A collection of SMTP addresses of users, distribution lists,
            or resources to get availability information for.
        """
        payload = {
            "schedules": schedules,
            "startTime": DateTimeTimeZone.parse(start_time),
            "endTime": DateTimeTimeZone.parse(end_time),
            "availabilityViewInterval": availability_view_interval
        }
        result = ClientResult(self.context, ClientValueCollection(ScheduleInformation))
        qry = ServiceOperationQuery(self, "getSchedule", None, payload, None, result)
        self.context.add_query(qry)
        return result

    @property
    def can_edit(self):
        """
        true if the user can write to the calendar, false otherwise.
        This property is true for the user who created the calendar.
        This property is also true for a user who has been shared a calendar and granted write access.

        :rtype: bool or None
        """
        return self.properties.get('canEdit', None)

    @property
    def name(self):
        """
        The calendar name.
        """
        return self.properties.get('name', None)

    @property
    def owner(self):
        """If set, this represents the user who created or added the calendar.
           For a calendar that the user created or added, the owner property is set to the user. For a calendar shared
           with the user, the owner property is set to the person who shared that calendar with the user.
        """
        return self.properties.get('owner', EmailAddress())

    @property
    def events(self):
        """The events in the calendar. Navigation property. Read-only."""
        return self.properties.get('events',
                                   EntityCollection(self.context, Event, ResourcePath("events", self.resource_path)))

    @property
    def calendar_view(self):
        """The calendar view for the calendar. Navigation property. Read-only.

        :rtype: EntityCollection
        """
        return self.get_property('calendarView',
                                 EntityCollection(self.context, Event,
                                                  ResourcePath("calendarView", self.resource_path)))

    @property
    def calendar_permissions(self):
        """The permissions of the users with whom the calendar is shared."""
        return self.properties.get('calendarPermissions',
                                   EntityCollection(self.context, CalendarPermission,
                                                    ResourcePath("calendarPermissions", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "calendarView": self.calendar_view,
                "calendarPermissions": self.calendar_permissions
            }
            default_value = property_mapping.get(name, None)
        return super(Calendar, self).get_property(name, default_value)



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.permissions.roleDefinitionCollection import RoleDefinitionCollection
from office365.sharepoint.principal.principal import Principal


class RoleAssignment(BaseEntity):
    """An association between a principal or a site group and a role definition."""

    @property
    def principal_id(self):
        """"""
        return self.properties.get("PrincipalId", None)

    @property
    def member(self):
        """Specifies the user or group corresponding to the role assignment."""
        return self.properties.get("Member",
                                   Principal(self.context, ResourcePath("Member", self.resource_path)))

    @property
    def role_definition_bindings(self):
        """Specifies a collection of role definitions for this role assignment."""
        return self.properties.get("RoleDefinitionBindings",
                                   RoleDefinitionCollection(self.context,
                                                            ResourcePath("RoleDefinitionBindings",
                                                                         self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "RoleDefinitionBindings": self.role_definition_bindings,
            }
            default_value = property_mapping.get(name, None)
        return super(RoleAssignment, self).get_property(name, default_value)



from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.folders.folder import Folder


class DocumentSet(Folder):

    @staticmethod
    def create(context, parent_folder, name, ct_id="0x0120D520"):
        """Creates a new DocumentSet object.

        :type context: office365.sharepoint.client_context.ClientContext
        :type parent_folder: office365.sharepoint.folders.folder.Folder
        :type name: str
        :type ct_id: office365.sharepoint.contenttypes.content_type_id.ContentTypeId
        """

        return_type = DocumentSet(context)

        def _get_list_details():
            custom_props = parent_folder.get_property("Properties")
            list_id = custom_props.get('vti_x005f_listname')
            target_list = context.web.lists.get_by_id(list_id)
            target_list.ensure_property("Title", _create_doc_set, target_list=target_list)

        def _create_doc_set(target_list):
            list_name = target_list.title.replace(" ", "")
            request_url = r"{0}/_vti_bin/listdata.svc/{1}".format(context.base_url, list_name)
            request = RequestOptions(request_url)
            request.method = HttpMethod.Post
            folder_url = parent_folder.serverRelativeUrl + '/' + name
            return_type._resource_path = ResourcePathServiceOperation("getFolderByServerRelativeUrl", [folder_url],
                                                                      ResourcePath("Web"))
            request.set_header('Slug', '{0}|{1}'.format(folder_url, ct_id))
            response = context.execute_request_direct(request)
            response.raise_for_status()
            json = response.json()
            context.pending_request().map_json(json, return_type)

        parent_folder.ensure_properties(["Properties", "ServerRelativeUrl"], _get_list_details)
        return return_type

    def set_property(self, name, value, persist_changes=True):
        super(Folder, self).set_property(name, value, persist_changes)
        return self



from office365.onenote.entity_base_model import OnenoteEntityBaseModel
from office365.runtime.client_result import ClientResult
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class OnenoteResource(OnenoteEntityBaseModel):
    """An image or other file resource on a OneNote page."""

    def get_content(self):
        """Retrieve the binary data of a file or image resource object."""
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "content", None, None, None, result)

        def _construct_query(request):
            """
            :type request: office365.runtime.http.request_options.RequestOptions
            """
            request.method = HttpMethod.Get
        self.context.before_execute(_construct_query)
        self.context.add_query(qry)
        return result

    @property
    def content_url(self):
        """The URL for downloading the content

        :rtype: str or None
        """
        return self.properties.get("contentUrl", None)



from office365.runtime.client_value import ClientValue


class LookupColumn(ClientValue):
    """
    The lookupColumn on a columnDefinition resource indicates that the column's values
    are looked up from another source in the site.
    """
    pass



from office365.runtime.client_value import ClientValue
from office365.sharepoint.permissions.base_permissions import BasePermissions


class RoleDefinitionCreationInformation(ClientValue):

    def __init__(self):
        """Contains properties that are used as parameters to initialize a role definition."""
        super(RoleDefinitionCreationInformation, self).__init__()
        self.Name = None
        self.Description = None
        self.BasePermissions = BasePermissions()



import os

from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.internal.create_file import create_file_query
from office365.sharepoint.internal.upload_session import UploadSessionQuery
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.files.file import File
from office365.sharepoint.files.file_creation_information import FileCreationInformation


class FileCollection(BaseEntityCollection):
    """Represents a collection of File resources."""

    def __init__(self, context, resource_path=None):
        super(FileCollection, self).__init__(context, File, resource_path)

    def get(self):
        """

        :rtype: FileCollection
        """
        return super(FileCollection, self).get()

    def create_upload_session(self, source_path, chunk_size, chunk_uploaded=None, *chunk_func_args):
        """Upload a file as multiple chunks

        :param str source_path: path where file to upload resides
        :param int chunk_size: upload chunk size (in bytes)
        :param (long)->None or None chunk_uploaded: uploaded event
        :param chunk_func_args: arguments to pass to chunk_uploaded function
        :return: office365.sharepoint.files.file.File
        """
        file_size = os.path.getsize(source_path)
        if file_size > chunk_size:
            qry = UploadSessionQuery(self, source_path, chunk_size, chunk_uploaded, chunk_func_args)
            self.context.add_query(qry)
            return qry.return_type
        else:
            with open(source_path, 'rb') as content_file:
                file_content = content_file.read()
            info = FileCreationInformation(url=os.path.basename(source_path), overwrite=True, content=file_content)
            qry = create_file_query(self, info)
            self.context.add_query(qry)
            return qry.return_type

    def add(self, file_creation_information):
        """Creates a File resource

        :type file_creation_information: office365.sharepoint.files.file_creation_information.FileCreationInformation
        :rtype: office365.sharepoint.files.file.File
        """
        qry = create_file_query(self, file_creation_information)
        self.context.add_query(qry)
        return qry.return_type

    def add_template_file(self, url_of_file, template_file_type):
        """Adds a ghosted file to an existing list or document library.

        :param int template_file_type: refer TemplateFileType enum
        :param str url_of_file: server relative url of a file
        """
        target_file = File(self.context)
        self.add_child(target_file)
        qry = ServiceOperationQuery(self,
                                    "addTemplateFile",
                                    {
                                        "urlOfFile": url_of_file,
                                        "templateFileType": template_file_type
                                    }, None, None, target_file)
        self.context.add_query(qry)
        return target_file

    def get_by_url(self, url):
        """Retrieve File object by url"""
        return File(self.context, ResourcePathServiceOperation("GetByUrl", [url], self.resource_path))

    def get_by_id(self, _id):
        """Gets the File with the specified ID."""
        return File(self.context, ResourcePathServiceOperation("getById", [_id], self.resource_path))



import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

if is_py2:
    from urlparse import urlparse
    from urllib import quote
    from urlparse import urljoin
    import pytz as timezone
    from email import message_from_string as message_from_bytes_or_string
    from __builtin__ import xrange as range_or_xrange
elif is_py3:
    from urllib.parse import urlparse
    from urllib.parse import quote
    from urllib.parse import urljoin
    from datetime import timezone
    from email import message_from_bytes as message_from_bytes_or_string
    from builtins import range as range_or_xrange


def message_as_bytes_or_string(message):
    if is_py2:
        return message.as_string()
    else:
        return message.as_bytes()


def is_string_type(value):
    if is_py2:
        return isinstance(value, basestring)
    else:
        return type(value) is str


def is_absolute_url(url):
    return bool(urlparse(url).netloc)



class RoleType:
    """Specifies the types of role definitions that are available for users and groups."""

    None_ = 0
    """The role definition has no rights on the site (2)."""

    Guest = 1
    """The role definition has limited right to view pages and specific page elements.
    This role is used to give users access to a particular page, list, or item in a list, without granting rights
    to view the entire site (2). Users cannot be added explicitly to the guest role; users who are given access to
    lists or document libraries by using permissions for a specific list are added automatically to the guest role.
    he guest role cannot be customized or deleted."""

    Reader = 2
    """
    The role definition has a right to view items, personalize Web Parts, use alerts, and create a top-level site.
    A reader can only read a site (2); the reader cannot add content. When a reader creates a site (2), the reader
    becomes the site (2) owner and a member of the administrator role for the new site (2).
    This does not affect the user's role membership for any other site (2).
    Rights included are CreateSSCSite, ViewListItems, and ViewPages.
    """

    Contributor = 3
    """
    The role definition has reader rights, and a right to add items, edit items, delete items, manage list permissions,
     manage personal views, personalize Web Part Pages, and browse directories. Includes all rights in the reader role,
     and AddDelPrivateWebParts, AddListItems, BrowseDirectories, CreatePersonalGroups, DeleteListItems, EditListItems,
     ManagePersonalViews, and UpdatePersonalWebParts roles. Contributors cannot create new lists or document libraries,
     but they can add content to existing lists and document libraries.
    """

    WebDesigner = 4
    """Has Contributor rights, plus rights to cancel check-out, delete items, manage lists, add and customize pages,
    define and apply themes and borders, and link style sheets. Includes all rights in the Contributor role, plus the
    following: AddAndCustomizePages, ApplyStyleSheets, ApplyThemeAndBorder, CancelCheckout, ManageLists. WebDesigners
    can modify the structure of the site and create new lists or document libraries. The value = 4."""

    Administrator = 5
    """Has all rights from other roles, plus rights to manage roles and view usage analysis data.
    Includes all rights in the WebDesigner role, plus the following: ManageListPermissions, ManageRoles, ManageSubwebs,
    ViewUsageData. The Administrator role cannot be customized or deleted, and must always contain at least one member.
    Members of the Administrator role always have access to, or can grant themselves access to, any item in the
    Web site. The value = 5."""

    Editor = 6
    """
        The role definition has reader rights, plus rights to Review items. Inclues all rights in the Reader role, plus
        the following: ReviewListItems. Reviewers cannot create new lists and document libraries, but they can add
        comments to existing list items or documents.
        """

    Reviewer = 7
    """The role definition is a special reader role definition with restricted rights, it has a right to view
        the content of the file but cannot download the file directly. Includes all rights in the Guest role, plus
        the following: ViewPages, ViewListItems and ViewItemsRequiresOpen."""

    RestrictedReader = 8

    System = 255
    """For SharePoint internal use only. System roles can not be deleted, nor modified."""



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity


class SiteIconManager(BaseEntity):

    def __init__(self, context):
        super(SiteIconManager, self).__init__(context, ResourcePath("SiteIconManager"))

    def get_site_logo(self, site_url, target=None, _type=None, return_type=None):
        payload = {
            "siteUrl": site_url,
            "target": target,
            "type": _type
        }
        if return_type is None:
            return_type = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "GetSiteLogo", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def set_site_logo(self, relative_logo_url, _type=None, aspect=None, focalx=None, focaly=None, isFocalPatch=None):
        payload = {
            "relativeLogoUrl": relative_logo_url,
            "type": _type,
            "aspect": aspect,
            "focalx": focalx,
            "focaly": focaly,
            "isFocalPatch": isFocalPatch
        }
        qry = ServiceOperationQuery(self, "SetSiteLogo", None, payload)
        self.context.add_query(qry)
        return self

    @property
    def entity_type_name(self):
        return "Microsoft.SharePoint.Portal.SiteIconManager"



import json

import requests
from requests.structures import CaseInsensitiveDict

from office365.runtime.client_request import ClientRequest
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions


class ODataV4BatchRequest(ClientRequest):
    """ JSON batch request """

    def __init__(self, context):
        super(ODataV4BatchRequest, self).__init__(context)

    def build_request(self):
        url = "{0}/$batch".format(self.context.service_root_url())
        request = RequestOptions(url)
        request.method = HttpMethod.Post
        request.ensure_header('Content-Type', "application/json")
        request.ensure_header('Accept', "application/json")
        request.data = self._prepare_payload()
        return request

    def process_response(self, batch_response):
        """Parses an HTTP response.

        :type batch_response: requests.Response
        """
        for query_id, resp in self._extract_response(batch_response):
            resp.raise_for_status()
            sub_qry = self.current_query.ordered_queries[query_id]
            self.context.pending_request().add_query(sub_qry)
            self.context.pending_request().process_response(resp)
        self.context.pending_request().clear()

    def _extract_response(self, batch_response):
        """
        type batch_response: requests.Response
        """
        json_responses = batch_response.json()
        for json_resp in json_responses["responses"]:
            resp = requests.Response()
            resp.status_code = int(json_resp['status'])
            resp.headers = CaseInsensitiveDict(json_resp['headers'])
            resp._content = json.dumps(json_resp["body"]).encode('utf-8')
            yield int(json_resp["id"]), resp

    def _prepare_payload(self):
        """
        Serializes a batch request body.
        """

        requests_json = []
        for qry in self.current_query.queries:
            request_id = str(len(requests_json))
            request = qry.build_request()
            requests_json.append(self._normalize_request(request, request_id))

        return {"requests": requests_json}

    def _normalize_request(self, request, _id, depends_on=None):
        """

        :type request: RequestOptions
        :type _id: str
        :type depends_on:  list[str] or None
        """
        allowed_props = ["id", "method", "headers", "url", "body"]
        request_json = dict((k, v) for k, v in vars(request).items() if v is not None and k in allowed_props)
        request_json["id"] = _id
        if depends_on is not None:
            request_json["dependsOn"] = depends_on
        request_json["url"] = request_json["url"].replace(self.context.service_root_url(), "")
        return request_json

    @property
    def current_query(self):
        """
        :rtype: office365.runtime.queries.batch_query.BatchQuery
        """
        return self._current_query



from office365.communications.onlinemeetings.online_meeting import OnlineMeeting
from office365.entity_collection import EntityCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class OnlineMeetingCollection(EntityCollection):

    def __init__(self, context, resource_path=None):
        super(OnlineMeetingCollection, self).__init__(context, OnlineMeeting, resource_path)

    def create_or_get(self, externalId=None, startDateTime=None, endDateTime=None, subject=None, participants=None,
                      chatInfo=None):
        """Create an onlineMeeting object with a custom specified external ID. If the external ID already exists,
        this API will return the onlineMeeting object with that external ID.

        :param str externalId: The external ID. A custom ID. (Required)
        :param datetime.datetime startDateTime: The meeting start time in UTC.
        :param datetime.datetime endDateTime: The meeting end time in UTC.
        :param str subject: The subject of the online meeting.
        :param list[MeetingParticipant] participants: The participants associated with the online meeting.
             This includes the organizer and the attendees.

        :param ChatInfo chatInfo:
        """
        return_type = OnlineMeeting(self.context)
        payload = {
            "externalId": externalId,
            "startDateTime": startDateTime,
            "endDateTime": endDateTime,
            "subject": subject,
            "chatInfo": chatInfo,
            "participants": participants
        }
        qry = ServiceOperationQuery(self, "createOrGet", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type



from office365.communications.calls.participant_info import ParticipantInfo
from office365.entity import Entity


class Participant(Entity):
    """Represents a participant in a call."""

    @property
    def info(self):
        """Information about the participant."""
        return self.properties.get("info", ParticipantInfo())



class PrincipalType:
    """Specifies the type of a principal."""

    def __init__(self):
        pass

    None_ = 0
    User = 1
    DistributionList = 2
    SecurityGroup = 4
    SharePointGroup = 8
    All = 15



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.sitedesigns.site_design_metadata import SiteDesignMetadata
from office365.sharepoint.sitedesigns.site_script_metadata import SiteScriptMetadata
from office365.sharepoint.sites.siteDesignPrincipal import SiteDesignPrincipal


class SiteScriptUtility(BaseEntity):
    """Use class to automate provisioning new or existing modern
    SharePoint sites that use your own custom configurations.
    """

    def __init__(self, context):
        super(SiteScriptUtility, self).__init__(context,
                                                ResourcePath(
                                                    "Microsoft.SharePoint.Utilities.WebTemplateExtensions.SiteScriptUtility"))

    @staticmethod
    def create_site_script(context, title, description, content):
        """
        Creates a new site script.

        :param office365.sharepoint.client_context.ClientContext context: SharePoint context
        :param str title:
        :param str description:
        :param str content:
        """
        return_type = ClientResult(context, SiteScriptMetadata)
        utility = SiteScriptUtility(context)
        payload = {
            "Title": title,
            "Description": description,
            "Content": content
        }
        qry = ServiceOperationQuery(utility, "CreateSiteScript", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def get_site_scripts(context, store=None):
        """
        Gets a list of information on all existing site scripts.

        :param office365.sharepoint.client_context.ClientContext context: SharePoint context
        :param str store:
        """
        return_type = ClientResult(context, ClientValueCollection(SiteScriptMetadata))
        utility = SiteScriptUtility(context)
        payload = {
            "store": store
        }
        qry = ServiceOperationQuery(utility, "GetSiteScripts", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def get_site_design_metadata(context, _id, store=None):
        """
        Gets information about a specific site script.

        :param office365.sharepoint.client_context.ClientContext context: SharePoint context
        :param str _id:
        :param str store:
        """
        return_type = ClientResult(context, SiteDesignMetadata())
        utility = SiteScriptUtility(context)
        payload = {
            "id": _id,
            "store": store
        }
        qry = ServiceOperationQuery(utility, "GetSiteDesignMetadata", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def get_site_design_rights(context, _id):
        """
        Gets a list of principals that have access to a site design.

        :type _id: str
        :param office365.sharepoint.client_context.ClientContext context: SharePoint client

        """
        return_type = BaseEntityCollection(context, SiteDesignPrincipal)
        utility = SiteScriptUtility(context)
        qry = ServiceOperationQuery(utility, "GetSiteDesignRights", [_id], None, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def grant_site_design_rights(context, _id, principal_names, granted_rights):
        """Grants access to a site design for one or more principals.

        :param office365.sharepoint.client_context.ClientContext context: SharePoint client
        :param str _id:
        :param list[str] principal_names:
        :param long granted_rights:
        """
        utility = SiteScriptUtility(context)
        payload = {
            "id": _id,
            "principalNames": ClientValueCollection(str, principal_names),
            "grantedRights": granted_rights
        }
        qry = ServiceOperationQuery(utility, "GrantSiteDesignRights", None, payload)
        qry.static = True
        context.add_query(qry)
        return utility

    @staticmethod
    def delete_site_design(context, _id):
        """
        Deletes a site design.

        :type _id: str
        :param office365.sharepoint.client_context.ClientContext context: SharePoint client
        """
        utility = SiteScriptUtility(context)
        qry = ServiceOperationQuery(utility, "DeleteSiteDesign", [_id])
        qry.static = True
        context.add_query(qry)
        return utility



from office365.runtime.client_value import ClientValue


class GroupProfile(ClientValue):
    def __init__(self, name, description=None, mail_enabled=False, security_enabled=True):
        """

        :param str name: Group name
        """
        super(GroupProfile, self).__init__()
        self.mailNickname = name
        self.displayName = name
        self.description = description
        self.mailEnabled = mail_enabled
        self.securityEnabled = security_enabled
        self.owners = []
        self.members = []
        self.groupTypes = []



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.permissions.role_definition import RoleDefinition
from office365.sharepoint.sharing.userSharingResult import UserSharingResult


class SharedWithMeViewItemRemovalResult(ClientValue):
    pass


class DocumentSharingManager(BaseEntity):
    """Specifies document sharing related methods."""

    def get_role_definition(self, role):
        """This method returns a role definition in the current web that is associated with a given Role
        (section 3.2.5.188) value.

        :param int role: A Role value for which to obtain the associated role definition object.
        """
        role_def = RoleDefinition(self.context)
        self.context.web.role_definitions.add_child(role_def)
        qry = ServiceOperationQuery(self, "GetRoleDefinition", [role], None, None, role_def)
        qry.static = True
        self.context.add_query(qry)
        return role_def

    def remove_items_from_shared_with_me_view(self, item_urls):
        """
        :type item_urls: list[str]
        """
        result = ClientResult(self.context, ClientValueCollection(SharedWithMeViewItemRemovalResult))
        qry = ServiceOperationQuery(self, "RemoveItemsFromSharedWithMeView", [item_urls], None, None, result)
        qry.static = True
        self.context.add_query(qry)
        return result

    def update_document_sharing_info(self, resourceAddress, userRoleAssignments, validateExistingPermissions,
                                     additiveMode, sendServerManagedNotification, customMessage,
                                     includeAnonymousLinksInNotification, propagateAcl):
        """

        :param str resourceAddress:
        :param ClientValueCollection userRoleAssignments:
        :param bool validateExistingPermissions:
        :param bool additiveMode:
        :param bool sendServerManagedNotification:
        :param str customMessage:
        :param bool includeAnonymousLinksInNotification:
        :param bool propagateAcl:
        """
        result = ClientResult(self.context, ClientValueCollection(UserSharingResult))
        payload = {
            "resourceAddress": resourceAddress,
            "userRoleAssignments": userRoleAssignments,
            "validateExistingPermissions": validateExistingPermissions,
            "additiveMode": additiveMode,
            "sendServerManagedNotification": sendServerManagedNotification,
            "customMessage": customMessage,
            "includeAnonymousLinksInNotification": includeAnonymousLinksInNotification,
            "propagateAcl": propagateAcl
        }
        qry = ServiceOperationQuery(self, "UpdateDocumentSharingInfo", None, payload, None, result)
        qry.static = True
        self.context.add_query(qry)
        return result

    @property
    def entity_type_name(self):
        return "SP.Sharing.DocumentSharingManager"



from office365.directory.certificates.certificate_authority import CertificateAuthority
from office365.entity import Entity
from office365.runtime.client_value_collection import ClientValueCollection


class CertificateBasedAuthConfiguration(Entity):

    @property
    def certificate_authorities(self):
        return self.properties.get("certificateAuthorities", ClientValueCollection(CertificateAuthority))



from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.changes.change_token import ChangeToken


class Change(BaseEntity):

    @property
    def change_token(self):
        """
        Returns an ChangeToken that represents the change.
        """
        return self.properties.get("ChangeToken", ChangeToken())

    @property
    def change_type(self):
        """
        Returns an SPChangeType that indicates the type of change, including adding, updating, deleting, or renaming
        changes, but also moving items away from or into lists and folders.
        """
        return self.properties.get("ChangeType", None)

    @property
    def site_id(self):
        """
        Returns the Id of the site of the changed item
        """
        return self.properties.get("SiteId", None)

    @property
    def time(self):
        """
        Gets a value that specifies the time that the object was modified.
        """
        return self.properties.get("Time", None)



import abc
from time import sleep

from office365.runtime.client_request_exception import ClientRequestException
from office365.runtime.compat import is_absolute_url
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.queries.read_entity_query import ReadEntityQuery


class ClientRuntimeContext(object):

    def build_request(self):
        return self.pending_request().build_request()

    def build_single_request(self, query):
        """

        :type: office365.runtime.queries.client_query.ClientQuery
        """
        return self.pending_request().build_single_request(query)

    def execute_query_retry(self, max_retry=5, timeout_secs=5, success_callback=None, failure_callback=None,
                            exceptions=(ClientRequestException,),
                            **kwargs):
        """
        Executes the current set of data retrieval queries and method invocations and retries it if needed.

        :param int max_retry: Number of times to retry the request
        :param int timeout_secs: Seconds to wait before retrying the request.
        :param (office365.runtime.client_object.ClientObject)-> None success_callback:
        :param (int, requests.exceptions.RequestException)-> None failure_callback:
        :param exceptions: tuple of exceptions that we retry
        """

        for retry in range(1, max_retry):
            try:
                self.execute_query(**kwargs)
                if callable(success_callback):
                    success_callback(self.current_query.return_type)
                break
            except exceptions as e:
                self.add_query(self.current_query, True)
                sleep(timeout_secs)
                if callable(failure_callback):
                    failure_callback(retry, e)

    @abc.abstractmethod
    def pending_request(self):
        """
        :rtype: office365.runtime.client_request.ClientRequest
        """
        pass

    @abc.abstractmethod
    def service_root_url(self):
        pass

    @abc.abstractmethod
    def authenticate_request(self, request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions
        """
        pass

    def load(self, client_object, properties_to_retrieve=None):
        """Prepare retrieval query

        :type properties_to_retrieve: list[str] or None
        :type client_object: office365.runtime.client_object.ClientObject
        """
        qry = ReadEntityQuery(client_object, properties_to_retrieve)
        self.add_query(qry)
        return qry

    def before_execute(self, action, once=True, *args, **kwargs):
        """
        Attach an event handler which is triggered before request is submitted to server

        :param (RequestOptions, any) -> None action:
        :param bool once:
        :return: None
        """

        def _process_request(request):
            if once:
                self.pending_request().beforeExecute -= _process_request
            action(request, *args, **kwargs)

        self.pending_request().beforeExecute += _process_request

    def after_query_execute(self, query, action, *args, **kwargs):
        """
        Attach an event handler which is triggered after query is submitted to server

        :type query: office365.runtime.queries.client_query.ClientQuery
        :type action: (Response, *args, **kwargs) -> None
        :return: None
        """

        def _process_response(resp):
            """
            :type resp: requests.Response
            """
            if self.current_query.id == query.id:
                action(*args, **kwargs)
                self.pending_request().afterExecute -= _process_response

        self.pending_request().afterExecute += _process_response

    def after_execute(self, action, once=True, *args, **kwargs):
        """
        Attach an event handler which is triggered after request is submitted to server

        :param (RequestOptions, *args, **kwargs) -> None action:
        :param bool once:
        :return: None
        """

        def _process_response(response):
            if once:
                self.pending_request().afterExecute -= _process_response
            action(response, *args, **kwargs)

        self.pending_request().afterExecute += _process_response

    def execute_request_direct(self, request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions or str
        """
        return self.pending_request().execute_request_direct(self._normalize_request(request))

    def execute_query(self,**kwargs):
        self.pending_request().execute_query(**kwargs)

    def add_query(self, query, execute_first=False, set_as_current=True):
        """
        :type query: office365.runtime.queries.client_query.ClientQuery
        :type execute_first: bool
        :type set_as_current: bool
        """
        self.pending_request().add_query(query, execute_first, set_as_current)

    def clear_queries(self):
        self.pending_request().queries.clear()

    @property
    def current_query(self):
        """
        :rtype: office365.runtime.queries.client_query.ClientQuery
        """
        return self.pending_request().current_query

    def _normalize_request(self, request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions or str
        """
        if not isinstance(request, RequestOptions):
            request = RequestOptions(request)

        if not is_absolute_url(request.url):
            url_parts = [self.service_root_url()]
            if not request.url.startswith("/"):
                url_parts.append("/")
            url_parts.append(request.url)
            request.url = "".join(url_parts)
        return request



from office365.entity import Entity


class OnlineMeeting(Entity):
    """
    Contains information about a meeting, including the URL used to join a meeting,
    the attendees list, and the description.
    """

    @property
    def subject(self):
        """The subject of the online meeting."""
        return self.properties.get("subject", None)

    @property
    def join_web_url(self):
        """The join URL of the online meeting. Read-only."""
        return self.properties.get("joinWebUrl", None)



from office365.directory.directory_object import DirectoryObject


class GroupSettingTemplate(DirectoryObject):
    """Group setting templates represent system-defined settings available to the tenant. Group settings can be
    created based on the available groupSettingTemplates, and values changed from their preset defaults. """



from office365.runtime.client_object import ClientObject


class SpoOperation(ClientObject):

    def __init__(self, context):
        super(SpoOperation, self).__init__(context)

    @property
    def is_complete(self):
        if self.is_property_available("IsComplete"):
            return bool(self.properties["IsComplete"])
        return None

    @property
    def polling_interval_secs(self):
        if self.is_property_available("PollingInterval"):
            return int(self.properties["PollingInterval"]) / 1000
        return None



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.alerts.alert import Alert
from office365.sharepoint.base_entity_collection import BaseEntityCollection


class AlertCollection(BaseEntityCollection):
    """Content Type resource collection"""

    def __init__(self, context, resource_path=None):
        super(AlertCollection, self).__init__(context, Alert, resource_path)

    def add(self, parameters):
        """

        :type parameters: office365.sharepoint.alerts.alert_creation_information.AlertCreationInformation
        """
        alert = Alert(self.context, None)
        self.add_child(alert)
        qry = ServiceOperationQuery(self, "Add", None, parameters, "alertCreationInformation", alert)
        self.context.add_query(qry)
        return alert

    def contains(self, id_alert):
        """
        :param str id_alert:
        """
        return_type = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "Contains", {"idAlert": id_alert}, None, None, return_type)
        self.context.add_query(qry)
        return return_type



from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class TaxonomyFieldValue(ClientValue):

    def __init__(self, label=None, term_guid=None, wss_id=-1):
        """
        :type label: str
        :type term_guid: str
        """
        super(TaxonomyFieldValue, self).__init__()
        self.Label = label
        self.TermGuid = term_guid
        self.WssId = wss_id

    def __str__(self):
        return "{0};#{1}|{2}".format(self.WssId, self.Label, self.TermGuid)

    @property
    def entity_type_name(self):
        return "SP.Taxonomy.TaxonomyFieldValue"


class TaxonomyFieldValueCollection(ClientValueCollection):
    """Represents the multi-value object for the taxonomy column."""

    def __init__(self, initial_values):
        super(TaxonomyFieldValueCollection, self).__init__(TaxonomyFieldValue, initial_values)

    def __str__(self):
        return ";#".join([str(item) for item in self._data])



from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class EmailProperties(ClientValue):

    def __init__(self, body, subject, to, from_address=None, cc=None, bcc=None, additional_headers=None):
        """

        :param str body:
        :param str subject:
        :param list[str] to:
        :param str or None from_address:
        :param list[str] or None cc:
        :param list[str] or None bcc:
        :param dict or None additional_headers:
        """
        super(EmailProperties, self).__init__()
        self.Body = body
        self.Subject = subject
        self.From = from_address
        self.To = ClientValueCollection(str, to)
        self.CC = ClientValueCollection(str, cc)
        self.BCC = ClientValueCollection(str, bcc)
        self.AdditionalHeaders = additional_headers

    @property
    def entity_type_name(self):
        return "SP.Utilities.EmailProperties"



import json
import re

import requests
from requests.structures import CaseInsensitiveDict

from office365.runtime.compat import message_from_bytes_or_string, message_as_bytes_or_string
from email.message import Message

from office365.runtime.client_request import ClientRequest
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.queries.batch_query import BatchQuery, create_boundary


class ODataBatchRequest(ClientRequest):

    def __init__(self, context):
        super(ODataBatchRequest, self).__init__(context)

    def build_request(self):
        """
        Construct a OData v3 Batch request
        """
        url = "{0}/$batch".format(self.context.service_root_url())
        request = RequestOptions(url)
        request.method = HttpMethod.Post
        media_type = "multipart/mixed"
        content_type = "; ".join([media_type, "boundary={0}".format(self.current_query.current_boundary)])
        request.ensure_header('Content-Type', content_type)
        request.data = self._prepare_payload()
        return request

    def process_response(self, batch_response):
        """
        Parses an HTTP response.

        :type batch_response: requests.Response
        """
        query_id = 0
        for response in self._extract_response(batch_response):
            response.raise_for_status()
            qry = self.current_query.ordered_queries[query_id]
            self.context.pending_request().add_query(qry, reset_queue=True)
            self.context.pending_request().process_response(response)
            query_id += 1
        self.context.pending_request().clear()

    def _extract_response(self, response):
        """Parses a multipart/mixed response body from from the position defined by the context.

        :type response: requests.Response
        """
        content_type = response.headers['Content-Type'].encode("ascii")
        http_body = (
            b"Content-Type: "
            + content_type
            + b"\r\n\r\n"
            + response.content
        )

        message = message_from_bytes_or_string(http_body)  # type: Message
        for raw_response in message.get_payload():
            if raw_response.get_content_type() == "application/http":
                yield self._deserialize_response(raw_response)

    def _prepare_payload(self):
        """
        Serializes a batch request body.
        """
        main_message = Message()
        main_message.add_header("Content-Type", "multipart/mixed")
        main_message.set_boundary(self.current_query.current_boundary)

        if self.current_query.has_change_sets:
            change_set_message = Message()
            change_set_boundary = create_boundary("changeset_", True)
            change_set_message.add_header("Content-Type", "multipart/mixed")
            change_set_message.set_boundary(change_set_boundary)

            for qry in self.current_query.change_sets:
                request = qry.build_request()
                message = self._serialize_request(request)
                change_set_message.attach(message)
            main_message.attach(change_set_message)

        for qry in self.current_query.get_queries:
            request = qry.build_request()
            message = self._serialize_request(request)
            main_message.attach(message)

        return message_as_bytes_or_string(main_message)

    @staticmethod
    def _normalize_headers(headers_raw):
        """
        :type headers_raw: list[str]
        """
        headers = {}
        for header_line in headers_raw:
            k, v = header_line.split(":", 1)
            headers[k.title()] = v.strip()
        return CaseInsensitiveDict(headers)

    def _deserialize_response(self, raw_response):
        """
        :type raw_response: Message
        """
        response = raw_response.get_payload(decode=True)
        lines = list(filter(None, response.decode("utf-8").split("\r\n")))
        response_status_regex = "^HTTP/1\\.\\d (\\d{3}) (.*)$"
        status_result = re.match(response_status_regex, lines[0])
        status_info = status_result.groups()

        resp = requests.Response()
        resp.status_code = int(status_info[0])
        if status_info[1] == "No Content" or len(lines) < 3:
            resp.headers = self._normalize_headers(lines[1:])
            resp._content = bytes(str("").encode("utf-8"))
        else:
            resp._content = bytes(str(lines[-1]).encode("utf-8"))
            resp.headers = self._normalize_headers(lines[1:-1])
        return resp

    @staticmethod
    def _serialize_request(request):
        """Serializes a part of a batch request to a string. A part can be either a GET request or
            a change set grouping several CUD (create, update, delete) requests.

        :type request: RequestOptions
        """
        eol = "\r\n"
        method = request.method
        if "X-HTTP-Method" in request.headers:
            method = request.headers["X-HTTP-Method"]
        lines = ["{method} {url} HTTP/1.1".format(method=method, url=request.url)] + \
                [':'.join(h) for h in request.headers.items()]
        if request.data:
            lines.append(eol)
            lines.append(json.dumps(request.data))
        raw_content = eol + eol.join(lines) + eol
        payload = raw_content.encode('utf-8').lstrip()

        message = Message()
        message.add_header("Content-Type", "application/http")
        message.add_header("Content-Transfer-Encoding", "binary")
        message.set_payload(payload)
        return message

    @property
    def current_query(self):
        """
        :rtype: BatchQuery
        """
        return self._current_query



from office365.entity import Entity
from office365.sharepoint.base_entity import BaseEntity


class GetExternalUsersResults(BaseEntity):
    pass


class RemoveExternalUsersResults(BaseEntity):
    pass


class SPOUserSessionRevocationResult(Entity):
    pass



class ChangeType:

    def __init__(self):
        pass

    NoChange = "NoChange"
    Add = "Add"
    Update = "Update"
    DeleteObject = "DeleteObject"
    Rename = "Rename"
    MoveAway = "MoveAway"
    MoveInto = "MoveInto"
    Restore = "Restore"
    RoleAdd = "RoleAdd"
    RoleDelete = "RoleDelete"
    RoleUpdate = "RoleUpdate"
    AssignmentAdd = "AssignmentAdd"
    AssignmentDelete = "AssignmentDelete"
    MemberAdd = "MemberAdd"
    MemberDelete = "MemberDelete"
    SystemUpdate = "SystemUpdate"
    Navigation = "Navigation"
    ScopeAdd = "ScopeAdd"
    ScopeDelete = "ScopeDelete"
    ListContentTypeAdd = "ListContentTypeAdd"
    ListContentTypeDelete = "ListContentTypeDelete"



from office365.runtime.client_result import ClientResult
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.files.file_version_event import FileVersionEvent
from office365.sharepoint.internal.download_file import create_download_file_query
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.directory.user import User
from office365.sharepoint.files.file_version_collection import FileVersionCollection
from office365.sharepoint.listitems.listitem import ListItem
from office365.sharepoint.permissions.information_rights_management_settings import InformationRightsManagementSettings
from office365.sharepoint.webparts.limited_webpart_manager import LimitedWebPartManager
from office365.sharepoint.types.resource_path import ResourcePath as SPResPath


class AbstractFile(BaseEntity):
    def read(self):
        """Immediately read content of file"""
        if not self.is_property_available("ServerRelativeUrl"):
            raise ValueError
        response = File.open_binary(
            self.context, self.properties["ServerRelativeUrl"])
        return response.content

    def write(self, content):
        """Immediately writes content of file"""
        if not self.is_property_available("ServerRelativeUrl"):
            raise ValueError
        response = File.save_binary(
            self.context, self.properties["ServerRelativeUrl"], content)
        return response


class File(AbstractFile):
    """Represents a file in a SharePoint Web site that can be a Web Part Page, an item in a document library,
    or a file in a folder."""

    @staticmethod
    def from_url(abs_url):
        """
        Retrieves a File from absolute url

        :type abs_url: str
        """
        from office365.sharepoint.client_context import ClientContext
        ctx = ClientContext.from_url(abs_url)
        file_relative_url = abs_url.replace(ctx.base_url, "")
        file = ctx.web.get_file_by_server_relative_url(file_relative_url)
        return file

    def get_image_preview_uri(self, width, height, client_type=None):
        """
        :param int width:
        :param int height:
        :param str client_type:
        """
        result = ClientResult(self.context)
        payload = {
            "width": width,
            "height": height,
            "clientType": client_type
        }
        qry = ServiceOperationQuery(self, "GetImagePreviewUri", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def get_image_preview_url(self, width, height, client_type=None):
        """
        :param int width:
        :param int height:
        :param str client_type:
        """
        result = ClientResult(self.context)
        payload = {
            "width": width,
            "height": height,
            "clientType": client_type
        }
        qry = ServiceOperationQuery(self, "GetImagePreviewUrl", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def recycle(self):
        """Moves the file to the Recycle Bin and returns the identifier of the new Recycle Bin item."""

        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "Recycle", None, None, None, result)
        self.context.add_query(qry)
        return result

    def approve(self, comment):
        """Approves the file submitted for content approval with the specified comment.

        :type comment: str
        """
        qry = ServiceOperationQuery(self,
                                    "approve",
                                    {
                                        "comment": comment
                                    })
        self.context.add_query(qry)
        return self

    def deny(self, comment):
        """Denies approval for a file that was submitted for content approval.

        :type comment: str
        """
        qry = ServiceOperationQuery(self,
                                    "deny",
                                    {
                                        "comment": comment
                                    })
        self.context.add_query(qry)
        return self

    def copyto(self, new_relative_url, overwrite):
        """Copies the file to the destination URL.

        :type new_relative_url: str
        :type overwrite: bool
        """
        qry = ServiceOperationQuery(self,
                                    "CopyTo",
                                    {
                                        "strNewUrl": new_relative_url,
                                        "boverwrite": overwrite
                                    },
                                    None)
        self.context.add_query(qry)
        return self

    def copyto_using_path(self, decoded_url, overwrite):
        """Copies the file to the destination URL.

        :type decoded_url: str
        :type overwrite: bool
        """
        qry = ServiceOperationQuery(self,
                                    "CopyToUsingPath",
                                    {
                                        "DecodedUrl": decoded_url,
                                        "bOverWrite": overwrite
                                    },
                                    None)
        self.context.add_query(qry)
        return self

    def moveto(self, new_relative_url, flag):
        """Moves the file to the specified destination URL.

        :type new_relative_url: str
        :type flag: int
        """
        qry = ServiceOperationQuery(self,
                                    "moveto",
                                    {
                                        "newurl": new_relative_url,
                                        "flags": flag
                                    },
                                    None)
        self.context.add_query(qry)
        return self

    def publish(self, comment):
        """Submits the file for content approval with the specified comment.
        :type comment: str
        """
        qry = ServiceOperationQuery(self,
                                    "publish",
                                    {
                                        "comment": comment,
                                    }
                                    )
        self.context.add_query(qry)

    def unpublish(self, comment):
        """Removes the file from content approval or unpublish a major version.
        :type comment: str
        """
        qry = ServiceOperationQuery(self,
                                    "unpublish",
                                    {
                                        "comment": comment,
                                    }
                                    )
        self.context.add_query(qry)
        return self

    def checkout(self):
        """Checks out the file from a document library based on the check-out type."""
        qry = ServiceOperationQuery(self,
                                    "checkout",
                                    )
        self.context.add_query(qry)
        return self

    def checkin(self, comment, checkin_type):
        """
        Checks the file in to a document library based on the check-in type.

        :param comment: comment to the new version of the file
        :param checkin_type: 0 (minor), or 1 (major) or 2 (overwrite)
            For more information on checkin types, please see
            https://docs.microsoft.com/en-us/previous-versions/office/sharepoint-csom/ee542953(v%3Doffice.15)
        :type checkin_type: int
        """
        qry = ServiceOperationQuery(self,
                                    "checkin",
                                    {
                                        "comment": comment,
                                        "checkInType": checkin_type
                                    }
                                    )
        self.context.add_query(qry)
        return self

    def undocheckout(self):
        """Reverts an existing checkout for the file."""
        qry = ServiceOperationQuery(self,
                                    "undocheckout"
                                    )
        self.context.add_query(qry)
        return self

    def get_limited_webpart_manager(self, scope):
        """Specifies the control set used to access, modify, or add Web Parts associated with this Web Part Page and
        view. """
        return LimitedWebPartManager(self.context,
                                     ResourcePathServiceOperation(
                                         "getlimitedwebpartmanager",
                                         [scope],
                                         self.resource_path
                                     ))

    def open_binary_stream(self):
        """Opens the file as a stream."""
        return_stream = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "OpenBinaryStream", None, None, None, return_stream)
        self.context.add_query(qry)
        return return_stream

    def save_binary_stream(self, stream):
        """Saves the file."""
        qry = ServiceOperationQuery(self, "SaveBinaryStream", None, {"file": stream})
        self.context.add_query(qry)
        return self

    def get_upload_status(self, upload_id):
        payload = {
            "uploadId": upload_id,
        }
        qry = ServiceOperationQuery(self, "GetUploadStatus", None, payload)
        self.context.add_query(qry)
        return self

    def upload_with_checksum(self, upload_id, checksum, stream):
        """
        :param str upload_id:
        :param str checksum:
        :param bytes stream:
        """
        return_type = File(self.context)
        payload = {
            "uploadId": upload_id,
            "checksum": checksum,
            "stream": stream
        }
        qry = ServiceOperationQuery(self, "UploadWithChecksum", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def cancel_upload(self, upload_id):
        payload = {
            "uploadId": upload_id,
        }
        qry = ServiceOperationQuery(self, "CancelUpload", None, payload)
        self.context.add_query(qry)
        return self

    def start_upload(self, upload_id, content):
        """Starts a new chunk upload session and uploads the first fragment.

        :param bytes content: File content
        :param str upload_id: Upload session id
        """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self,
                                    "startUpload",
                                    {
                                        "uploadID": upload_id
                                    },
                                    content,
                                    None,
                                    result
                                    )
        self.context.add_query(qry)
        return result

    def continue_upload(self, upload_id, file_offset, content):
        """
        Continues the chunk upload session with an additional fragment. The current file content is not changed.

        :param str upload_id: Upload session id
        :param int file_offset: File offset
        :param bytes content: File content
        """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self,
                                    "continueUpload",
                                    {
                                        "uploadID": upload_id,
                                        "fileOffset": file_offset,
                                    },
                                    content,
                                    None,
                                    result
                                    )
        self.context.add_query(qry)
        return result

    def finish_upload(self, upload_id, file_offset, content):
        """Uploads the last file fragment and commits the file. The current file content is changed when this method
        completes.

        :param str upload_id: Upload session id
        :param int file_offset: File offset
        :param bytes content: File content
        """
        qry = ServiceOperationQuery(self,
                                    "finishUpload",
                                    {
                                        "uploadID": upload_id,
                                        "fileOffset": file_offset,
                                    },
                                    content,
                                    None,
                                    self
                                    )
        self.context.add_query(qry)
        return self

    @staticmethod
    def save_binary(ctx, server_relative_url, content):
        """Uploads a file

        :type ctx: ClientContext
        :type server_relative_url: str
        :type content: str
        """
        url = r"web/getFileByServerRelativePath(DecodedUrl='{0}')/\$value".format(server_relative_url)
        request = RequestOptions(url)
        request.method = HttpMethod.Post
        request.set_header('X-HTTP-Method', 'PUT')
        request.data = content
        response = ctx.execute_request_direct(request)
        return response

    @staticmethod
    def open_binary(ctx, server_relative_url):
        """
        Returns the file object located at the specified server-relative URL.

        :type ctx: ClientContext
        :type server_relative_url: str
        :return Response
        """
        url = r"web/getFileByServerRelativePath(DecodedUrl='{0}')/\$value".format(server_relative_url)
        request = RequestOptions(url)
        request.method = HttpMethod.Get
        response = ctx.execute_request_direct(request)
        return response

    def download(self, file_object):
        """
        Download a file content

        :type file_object: typing.IO
        """

        def _download_inner():
            qry = create_download_file_query(self, file_object)
            self.context.add_query(qry)

        self.ensure_property("ServerRelativePath", _download_inner)
        return self

    def download_session(self, file_object, chunk_downloaded=None, chunk_size=1024 * 1024):
        """
        :type file_object: typing.IO
        :type chunk_downloaded: (int)->None or None
        :type chunk_size: int
        """

        def _download_as_stream():
            qry = ServiceOperationQuery(self, "$value")

            def _construct_download_request(request):
                """
                :type request: office365.runtime.http.request_options.RequestOptions
                """
                request.stream = True
                request.method = HttpMethod.Get

            self.context.before_execute(_construct_download_request)

            def _process_download_response(response):
                """
                :type response: requests.Response
                """
                response.raise_for_status()
                bytes_read = 0
                for chunk in response.iter_content(chunk_size=chunk_size):
                    bytes_read += len(chunk)
                    if callable(chunk_downloaded):
                        chunk_downloaded(bytes_read)
                    file_object.write(chunk)

            self.context.after_execute(_process_download_response)

            self.context.add_query(qry)

        self.ensure_property("ServerRelativeUrl", _download_as_stream)
        return self

    @property
    def checked_out_by_user(self):
        """Gets an object that represents the user who has checked out the file."""
        return self.properties.get('CheckedOutByUser',
                                   User(self.context, ResourcePath("CheckedOutByUser", self.resource_path)))

    @property
    def version_events(self):
        return self.properties.get("VersionEvents",
                                   BaseEntityCollection(self.context,
                                                        FileVersionEvent,
                                                        ResourcePath("VersionEvents", self.resource_path)))

    @property
    def information_rights_management_settings(self):
        return self.properties.get('InformationRightsManagementSettings',
                                   InformationRightsManagementSettings(self.context,
                                                                       ResourcePath(
                                                                           "InformationRightsManagementSettings",
                                                                           self.resource_path)))

    @property
    def listItemAllFields(self):
        """Gets a value that specifies the list item fields values for the list item corresponding to the file."""
        return self.properties.get('ListItemAllFields',
                                   ListItem(self.context, ResourcePath("listItemAllFields", self.resource_path)))

    @property
    def versions(self):
        """Gets a value that returns a collection of file version objects that represent the versions of the file."""
        return self.properties.get('Versions',
                                   FileVersionCollection(self.context, ResourcePath("versions", self.resource_path)))

    @property
    def modified_by(self):
        """
        Gets a value that returns the user who last modified the file.

        :rtype: office365.sharepoint.directory.user.User or None
        """
        return self.properties.get("ModifiedBy", User(self.context, ResourcePath("ModifiedBy", self.resource_path)))

    @property
    def locked_by_user(self):
        """
        Gets a value that returns the user that owns the current lock on the file.

        :rtype: office365.sharepoint.directory.user.User or None
        """
        return self.properties.get("LockedByUser", User(self.context, ResourcePath("LockedByUser", self.resource_path)))

    @property
    def serverRelativeUrl(self):
        """Gets the relative URL of the file based on the URL for the server.

        :rtype: str or None
        """
        return self.properties.get("ServerRelativeUrl", None)

    @property
    def server_relative_path(self):
        """Gets the server-relative Path of the list folder.
        :rtype: SPResPath or None
        """
        return self.properties.get("ServerRelativePath", SPResPath())

    @property
    def length(self):
        """Gets the file size.

        :rtype: int or None
        """
        return int(self.properties.get("Length", -1))

    @property
    def exists(self):
        """Specifies whether the file exists.

        :rtype: bool or None
        """
        return self.properties.get("Exists", None)

    @property
    def name(self):
        """Specifies the file name including the extension.
            It MUST NOT be NULL. Its length MUST be equal to or less than 260.

        :rtype: str or None
        """
        return self.properties.get("Name", None)

    @property
    def list_id(self):
        """Gets the GUID that identifies the List containing the file.

        :rtype: str or None
        """
        return self.properties.get("ListId", None)

    @property
    def site_id(self):
        """Gets the GUID that identifies the site collection containing the file.

        :rtype: str or None
        """
        return self.properties.get("SiteId", None)

    @property
    def web_id(self):
        """Gets the GUID for the site containing the file.

        :rtype: str or None
        """
        return self.properties.get("WebId", None)

    @property
    def time_created(self):
        """Gets a value that specifies when the file was created.

        :rtype: str or None
        """
        return self.properties.get("TimeCreated", None)

    @property
    def time_last_modified(self):
        """Specifies when the file was last modified.

        :rtype: str or None
        """
        return self.properties.get("TimeLastModified", None)

    @property
    def minor_version(self):
        """
        Gets a value that specifies the minor version of the file.
        """
        return int(self.properties.get("MinorVersion", -1))

    @property
    def major_version(self):
        """
        Gets a value that specifies the major version of the file.

        :rtype: int or None
        """
        return int(self.properties.get("MajorVersion", -1))

    @property
    def unique_id(self):
        """
        Gets a value that specifies the a file unique identifier

        :rtype: str or None
        """
        return self.properties.get("UniqueId", None)

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "CheckedOutByUser": self.checked_out_by_user,
                "VersionEvents": self.version_events,
                "InformationRightsManagementSettings": self.information_rights_management_settings,
                "LockedByUser": self.locked_by_user,
                "ModifiedBy": self.modified_by
            }
            default_value = property_mapping.get(name, None)
        return super(File, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(File, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if self._resource_path is None:
            if name == "ServerRelativeUrl":
                self._resource_path = ResourcePathServiceOperation("GetFileByServerRelativeUrl", [value],
                                                                   ResourcePath("Web"))
            elif name == "ServerRelativePath":
                self._resource_path = ResourcePathServiceOperation("getFolderByServerRelativePath", [value],
                                                                   ResourcePath("Web"))
            elif name == "UniqueId":
                self._resource_path = ResourcePathServiceOperation("GetFileById", [value], ResourcePath("Web"))
        return self



from office365.directory.identities.identity_set import IdentitySet
from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.onedrive.drives.drive_recipient import DriveRecipient
from office365.onedrive.listitems.item_reference import ItemReference
from office365.onedrive.permissions.sharing_invitation import SharingInvitation
from office365.onedrive.permissions.sharing_link import SharingLink
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class Permission(Entity):
    """The Permission resource provides information about a sharing permission granted for a DriveItem resource."""

    def grant(self, recipients, roles):
        """
        Grant users access to a link represented by a permission.

        :param list[str] recipients: A collection of recipients who will receive access.
        :param list[str] roles: If the link is an "existing access" link, specifies roles to be granted to the users.
            Otherwise must match the role of the link.
        """
        payload = {
            "recipients": ClientValueCollection(DriveRecipient, [DriveRecipient.from_email(r) for r in recipients]),
            "roles": ClientValueCollection(str, roles)
        }
        return_type = EntityCollection(self.context, Permission, ResourcePath("permissions", self.resource_path))
        qry = ServiceOperationQuery(self, "grant", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    @property
    def invitation(self):
        """For user type permissions, the details of the users & applications for this permission."""
        return self.properties.get('invitation', SharingInvitation())

    @property
    def granted_to(self):
        """For user type permissions, the details of the users & applications for this permission."""
        return self.properties.get('grantedTo', IdentitySet())

    @property
    def granted_to_identities(self):
        """For link type permissions, the details of the users to whom permission was granted. Read-only."""
        return self.properties.get('grantedToIdentities', ClientValueCollection(IdentitySet))

    @property
    def link(self):
        """Provides the link details of the current permission, if it is a link type permissions. Read-only."""
        return self.properties.get('link', SharingLink())

    @property
    def roles(self):
        """The type of permission, e.g. read. See below for the full list of roles. Read-only."""
        return self.properties.get('roles', ClientValueCollection(str))

    @roles.setter
    def roles(self, value):
        """
        Sets the type of permission

        :type value: list[str]
        """
        self.set_property("roles", ClientValueCollection(str, value))

    @property
    def share_id(self):
        """A unique token that can be used to access this shared item via the shares API. Read-only.

        :rtype: str
        """
        return self.properties.get('shareId', None)

    @property
    def has_password(self):
        """This indicates whether password is set for this permission, it's only showing in response.
        Optional and Read-only and for OneDrive Personal only.

        :rtype: bool
        """
        return self.properties.get('hasPassword', None)

    @property
    def inherited_from(self):
        """
        If this content type is inherited from another scope (like a site),
        provides a reference to the item where the content type is defined.
        """
        return self.properties.get("inheritedFrom", ItemReference())

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "inheritedFrom": self.inherited_from,
                "grantedTo": self.granted_to,
                "grantedToIdentities": self.granted_to_identities
            }
            default_value = property_mapping.get(name, None)
        return super(Permission, self).get_property(name, default_value)



from office365.runtime.auth.authentication_provider import AuthenticationProvider

try:
    from requests_ntlm import HttpNtlmAuth
except ImportError:
    raise ImportError("To use NTLM authentication the package 'requests_ntlm' needs to be installed.")


class NtlmProvider(AuthenticationProvider):

    def __init__(self, username, password):
        """
            Provides NTLM authentication (intended for SharePoint On-Premises)

            Note: due to Outlook REST API v1.0 BasicAuth Deprecation
            (refer https://developer.microsoft.com/en-us/office/blogs/outlook-rest-api-v1-0-basicauth-deprecation/)
            NetworkCredentialContext class should be no longer utilized for Outlook REST API v1.0

            :type username: str
            :type password: str
        """
        super(NtlmProvider, self).__init__()
        self.auth = HttpNtlmAuth(username, password)

    def authenticate_request(self, request):
        request.auth = self.auth



from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class UserSharingResult(ClientValue):

    def __init__(self):
        super(UserSharingResult, self).__init__()
        self.AllowedRoles = ClientValueCollection(int)
        self.CurrentRole = None
        self.DisplayName = None
        self.Email = None
        self.InvitationLink = None
        self.IsUserKnown = None
        self.Message = None
        self.Status = None
        self.User = None

    @property
    def entity_type_name(self):
        return "SP.Sharing"



from office365.runtime.client_value_collection import ClientValueCollection
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.sites.language import Language


class LanguageCollection(BaseEntity):

    def __init__(self, context, resource_path=None):
        super(LanguageCollection, self).__init__(context, resource_path)
        self.properties["Items"] = ClientValueCollection(Language)

    @property
    def items(self):
        """
        :rtype: ClientValueCollection
        """
        return self.properties.get("Items")

    def set_property(self, name, value, persist_changes=True):
        self.items.set_property(name, value, persist_changes)
        return self



from office365.sharepoint.fields.field import Field


class FieldText(Field):

    def __init__(self, context):
        super(FieldText, self).__init__(context)

    @property
    def max_length(self):
        """Gets a value that specifies the maximum number of characters allowed in the value of the field."""
        return self.properties.get('MaxLength', None)

    @max_length.setter
    def max_length(self, val):
        self.set_property("MaxLength", val, True)



from office365.runtime.client_object import ClientObject
from office365.runtime.odata.query_options import QueryOptions
from office365.runtime.queries.client_query import ClientQuery


class ReadEntityQuery(ClientQuery):
    def __init__(self, entity, properties_to_include=None):
        """
        Read entity query

        :type properties_to_include: list[str] or None
        :type entity: office365.runtime.client_object.ClientObject
        """
        super(ReadEntityQuery, self).__init__(entity.context, entity, None, None, entity)
        self._properties_to_include = properties_to_include

    def build_url(self):
        url = super(ReadEntityQuery, self).build_url()
        if self._properties_to_include:
            url += "?" + self._build_query_url()
        return url

    def _build_query_url(self):
        query = QueryOptions()
        for n in self._properties_to_include:
            prop_val = self._binding_type.get_property(n)
            if isinstance(prop_val, ClientObject) or n == "Properties":
                query.expand.append(n)
            query.select.append(n)
        return query.to_url()



from office365.runtime.client_object import ClientObject
from office365.runtime.client_result import ClientResult
from office365.runtime.http.request_options import RequestOptions
from office365.runtime.types.EventHandler import EventHandler


class ClientObjectCollection(ClientObject):

    def __init__(self, context, child_item_type, resource_path=None):
        """Client object collection

        :type context: office365.runtime.client_runtime_context.ClientRuntimeContext
        :type child_item_type: type[ClientObject]
        :type resource_path: office365.runtime.resource_path.ResourcePath
        """
        super(ClientObjectCollection, self).__init__(context, resource_path)
        self._data = []
        self._item_type = child_item_type
        self.page_loaded = EventHandler(False)
        self._default_page_size = 100
        self._paged_mode = True
        self._page_index = 0
        self.next_request_url = None

    @property
    def entity_type_name(self):
        name = super(ClientObjectCollection, self).entity_type_name
        return "Collection({0})".format(name)

    def get(self):
        """
        :rtype: ClientObjectCollection
        """
        return super(ClientObjectCollection, self).get()

    def clear(self):
        self._data = []

    def create_typed_object(self, properties=None, persist_changes=False):
        """
        :type properties: dict
        :type persist_changes: bool
        :rtype: ClientObject
        """
        if properties is None:
            properties = {}
        if self._item_type is None:
            raise AttributeError("No class for object type '{0}' found".format(self._item_type))

        client_object = self._item_type(self.context)
        client_object._parent_collection = self
        for k, v in properties.items():
            client_object.set_property(k, v, persist_changes)
        return client_object

    def set_property(self, name, value, persist_changes=False):
        """
        :type name: str
        :type value: any
        :type persist_changes: bool
        """
        child_client_object = self.create_typed_object(value)
        self.add_child(child_client_object)

    def add_child(self, client_object):
        """
        Adds client object into collection

        :type client_object: ClientObject
        """
        client_object._parent_collection = self
        self._data.append(client_object)

    def remove_child(self, client_object):
        """
        :type client_object: ClientObject
        """
        self._data = [item for item in self._data if item != client_object]

    def __iter__(self):
        """
        :rtype: collections.Iterable[ClientObject]
        """
        for cur_item in self._data:
            yield cur_item

        if self._paged_mode:
            while self.next_request_url:
                for cur_item in self._get_next_items():
                    yield cur_item

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        """

        :type index: int
        :rtype: ClientObject
        """
        item_iterator = iter(self)
        while len(self._data) <= index:
            next(item_iterator)
        return self._data[index]

    def to_json(self, json_format=None):
        return [item.to_json(json_format) for item in self._data]

    def filter(self, expression):
        """
        Sets OData $filter query option

        :type expression: str
        """
        self.query_options.filter = expression
        return self

    def order_by(self, value):
        """
        Sets OData $orderBy query option

        :type value: int
        """
        self.query_options.orderBy = value
        return self

    def skip(self, value):
        """
        Sets $skip system query option

        :type value: int
        """
        self.query_options.skip = value
        return self

    def top(self, value):
        """
        :type value: int
        """
        self._paged_mode = False
        self.query_options.top = value
        return self

    def get_items_count(self):
        """
        Gets total items count
        :return: ClientResult
        """
        result = ClientResult(self.context)

        def _calc_items_count(resp):
            list(iter(self))
            result.value = len(self)

        self.context.load(self)
        self.context.after_execute(_calc_items_count)
        return result

    def _load_paged_items(self):
        request = RequestOptions(self.next_request_url)
        response = self.context.execute_request_direct(request)
        json = response.json()
        self.next_request_url = None
        self.context.pending_request().map_json(json, self)

    def _get_next_items(self):
        self._page_index += 1
        next_index = self._default_page_size * self._page_index
        self._load_paged_items()
        self.page_loaded.notify(len(self._data))
        return self._data[next_index:]



from office365.runtime.client_value import ClientValue


class MeetingTimeSuggestion(ClientValue):
    """
    A meeting suggestion that includes information like meeting time, attendance likelihood, individual availability,
    and available meeting locations.
    """
    pass



from office365.directory.identities.identity_provider_base import IdentityProviderBase


class BuiltInIdentityProvider(IdentityProviderBase):
    """
    Represents built-in identity providers with External Identities for an Azure Active Directory tenant.

    """
    pass



from office365.runtime.client_value import ClientValue


class GeoCoordinates(ClientValue):
    """
    The GeoCoordinates resource provides geographic coordinates and elevation of a location based on metadata
    contained within the file. If a DriveItem has a non-null location facet, the item represents a file with
    a known location associated with it.
    """
    pass



from office365.runtime.client_value import ClientValue


class SubscriptionInformation(ClientValue):

    def __init__(self, notification_url=None, resource=None):
        super(SubscriptionInformation, self).__init__()
        self.notificationUrl = notification_url
        self.resource = resource



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.navigation.navigation_node_collection import NavigationNodeCollection


class Navigation(BaseEntity):
    """Represents navigation operations at the site collection level."""

    @property
    def use_shared(self):
        """Gets a value that specifies whether the site inherits navigation."""
        return self.properties.get('UseShared', None)

    @use_shared.setter
    def use_shared(self, value):
        """Sets a value that specifies whether the site inherits navigation."""
        self.set_property('UseShared', value)

    @property
    def quick_launch(self):
        """Gets a value that collects navigation nodes corresponding to links in the Quick Launch area of the site."""
        return self.properties.get('QuickLaunch',
                                   NavigationNodeCollection(self.context,
                                                            ResourcePath("QuickLaunch", self.resource_path)))

    @property
    def top_navigation_bar(self):
        """Gets a value that collects navigation nodes corresponding to links in the top navigation bar of the site."""
        return self.properties.get('TopNavigationBar',
                                   NavigationNodeCollection(self.context,
                                                            ResourcePath("TopNavigationBar", self.resource_path)))

    def set_property(self, name, value, persist_changes=True):
        super(Navigation, self).set_property(name, value, persist_changes)

    def get_property(self, name, default_value=None):
        if name == "TopNavigationBar":
            default_value = self.top_navigation_bar
        elif name == "QuickLaunch":
            default_value = self.quick_launch
        return super(Navigation, self).get_property(name, default_value)



from office365.runtime.client_value import ClientValue


class SharingInvitation(ClientValue):
    """The SharingInvitation resource groups invitation-related data items into a single structure."""
    pass



from office365.entity import Entity
from office365.onedrive.driveitems.thumbnail import Thumbnail


class ThumbnailSet(Entity):
    """The ThumbnailSet resource is a keyed collection of thumbnail resources.
    It is used to represent a set of thumbnails associated with a DriveItem."""

    @property
    def large(self):
        """A 1920x1920 scaled thumbnail."""
        return self.properties.get('drive', Thumbnail())

    @property
    def medium(self):
        """A 176x176 scaled thumbnail."""
        return self.properties.get('medium', Thumbnail())

    @property
    def small(self):
        """A 48x48 cropped thumbnail."""
        return self.properties.get('small', Thumbnail())

    @property
    def source(self):
        """A custom thumbnail image or the original image used to generate other thumbnails."""
        return self.properties.get('source', Thumbnail())



from office365.onedrive.internal.resourcepaths.resource_path_url import ResourcePathUrl
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


def create_upload_content_query(folder_item, name, content=None):
    """

    :param office365.onedrive.driveItem.DriveItem folder_item: Folder (container)
    :param str name: a file name
    :param str content: a file content
    """
    from office365.onedrive.driveitems.driveItem import DriveItem
    file_item = DriveItem(folder_item.context, ResourcePathUrl(name, folder_item.resource_path))
    qry = ServiceOperationQuery(file_item, "content", None, content, None, file_item)

    def _modify_query(request):
        """
        :type request: office365.runtime.http.request_options.RequestOptions
        """
        request.method = HttpMethod.Put
    folder_item.context.before_execute(_modify_query)
    return qry



from office365.runtime.client_value import ClientValue


class PersonOrGroupColumn(ClientValue):
    """The personOrGroupColumn on a columnDefinition resource indicates that the column's values represent
    a person or group chosen from the directory."""
    pass



from office365.runtime.client_value import ClientValue


class SPInvitationCreationResult(ClientValue):

    def __init__(self):
        super(SPInvitationCreationResult, self).__init__()
        self.Email = None
        self.InvitationLink = None
        self.Succeeded = None



from office365.directory.identities.identity_provider_base import IdentityProviderBase


class SocialIdentityProvider(IdentityProviderBase):
    """
    Represents social identity providers with External Identities for both Azure Active Directory (Azure AD)
    tenant and an Azure AD B2C tenant.
    """

    @property
    def client_id(self):
        """
        The client identifier for the application obtained when registering the application with the identity provider.

        :rtype: str or None
        """
        return self.properties.get('clientId', None)

    @property
    def client_secret(self):
        """
        The client secret for the application that is obtained when the application is registered
        with the identity provider. This is write-only. A read operation returns ****.

        :rtype: str or None
        """
        return self.properties.get('clientSecret', None)

    @property
    def identity_provider_type(self):
        """
        For a B2B scenario, possible values: Google, Facebook.
        For a B2C scenario, possible values: Microsoft, Google, Amazon, LinkedIn, Facebook, GitHub, Twitter, Weibo,
        QQ, WeChat.

        :rtype: str or None
        """
        return self.properties.get('identityProviderType', None)




from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.recyclebin.recycleBinItem import RecycleBinItem


class RecycleBinItemCollection(BaseEntityCollection):
    """Represents a collection of View resources."""

    def __init__(self, context, resource_path=None):
        super(RecycleBinItemCollection, self).__init__(context, RecycleBinItem, resource_path)

    def move_all_to_second_stage(self):
        qry = ServiceOperationQuery(self, "MoveAllToSecondStage")
        self.context.add_query(qry)
        return self

    def get_by_id(self, recycleBinId):
        """
        Returns the recycle bin type with the given identifier from the collection.

        :param str recycleBinId: A hexadecimal value representing the identifier of a recycle bin.
        """
        return RecycleBinItem(self.context,
                              ResourcePathServiceOperation("GetById", [recycleBinId], self.resource_path))

    def delete_all(self):
        """Permanently deletes all Recycle Bin items."""
        qry = ServiceOperationQuery(self, "DeleteAll")
        self.context.add_query(qry)
        return self

    def restore_all(self):
        """Restores all Recycle Bin items to their original locations."""
        qry = ServiceOperationQuery(self, "RestoreAll")
        self.context.add_query(qry)
        return self



from office365.entity import Entity
from office365.entity_collection import EntityCollection


class WorkbookWorksheet(Entity):
    """
    An Excel worksheet is a grid of cells. It can contain data, tables, charts, etc
    """

    @property
    def name(self):
        """
        The display name of the worksheet.
        :rtype: str or None
        """
        return self.properties.get('name', None)


class WorkbookWorksheetCollection(EntityCollection):

    def __init__(self, context, resource_path=None):
        super(WorkbookWorksheetCollection, self).__init__(context, WorkbookWorksheet, resource_path)



from office365.entity_collection import EntityCollection
from office365.onedrive.sites.site import Site
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class SitesWithRoot(EntityCollection):
    """Sites container"""

    def __init__(self, context, resource_path):
        super(SitesWithRoot, self).__init__(context, Site, resource_path)

    def remove(self, sites):
        """
        :type sites: SitesWithRoot
        """
        return_type = SitesWithRoot(self.context, self.resource_path)

        payload = {
            "value": sites,
        }
        qry = ServiceOperationQuery(self, "remove", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    @property
    def root(self):
        return self.properties.get('root',
                                   Site(self.context, ResourcePath("root", self.resource_path)))




from office365.runtime.client_object import ClientObject
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation


class UserProfilePropertiesForUser(ClientObject):

    def __init__(self, context, account_name, property_names):
        params = {
            "accountName": account_name,
            "propertyNames": property_names
        }

        super(UserProfilePropertiesForUser, self).__init__(context, ResourcePathServiceOperation("SP.UserProfiles.UserProfilePropertiesForUser", params))



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity


class Alert(BaseEntity):

    @property
    def user(self):
        """Gets user object that represents User for the alert."""
        from office365.sharepoint.principal.user import User
        return self.properties.get('User',
                                   User(self.context, ResourcePath("user", self.resource_path)))

    @property
    def list(self):
        """Gets list object that represents List for the alert."""
        from office365.sharepoint.lists.list import List
        return self.properties.get('List',
                                   List(self.context, ResourcePath("list", self.resource_path)))



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.userprofiles.personalSiteCreationPriority import PersonalSiteCreationPriority
from office365.sharepoint.userprofiles.personProperties import PersonProperties
from office365.sharepoint.userprofiles.personPropertiesCollection import PersonPropertiesCollection


class PeopleManager(BaseEntity):
    """Provides methods for operations related to people."""

    def __init__(self, context):
        super(PeopleManager, self).__init__(context, ResourcePath("SP.UserProfiles.PeopleManager"))

    def am_i_following(self, account_name):
        """
        Checks whether the current user is following the specified user.

        :param str account_name:
        :return:
        """
        result = ClientResult(self.context)
        params = {"accountName": account_name}
        qry = ServiceOperationQuery(self, "AmIFollowing", params, None, None, result)
        self.context.add_query(qry)
        return result

    def get_followers_for(self, account_name):
        """
        Gets the people who are following the specified user.

        :param str account_name:
        :return:
        """
        result = ClientResult(self.context, PersonPropertiesCollection(self.context))
        params = {"accountName": account_name}
        qry = ServiceOperationQuery(self, "GetFollowersFor", params, None, None, result)
        self.context.add_query(qry)
        return result

    def follow(self, account_name):
        """
        Add the specified user to the current user's list of followed users.

        :param str account_name:
        """
        params = {"accountName": account_name}
        qry = ServiceOperationQuery(self, "Follow", params, None, None, None)
        self.context.add_query(qry)
        return self

    def stop_following(self, account_name):
        """
        Remove the specified user from the current user's list of followed users.

        :param str account_name:
        """
        params = {"accountName": account_name}
        qry = ServiceOperationQuery(self, "StopFollowing", params, None, None, None)
        self.context.add_query(qry)
        return self

    def get_user_profile_properties(self, accountName):
        """
        Gets the specified user profile properties for the specified user.

        :type accountName: str
        :return: dict
        """
        result = ClientResult(self.context)
        payload = {"accountName": accountName}
        qry = ServiceOperationQuery(self, "GetUserProfileProperties", payload, None, None, result)
        self.context.add_query(qry)
        return result

    def get_properties_for(self, accountName):
        """
        Gets user properties for the specified user.

        :type accountName: str
        :return: PersonProperties
        """
        result = PersonProperties(self.context)
        payload = {"accountName": accountName}
        qry = ServiceOperationQuery(self, "GetPropertiesFor", payload, None, None, result)
        self.context.add_query(qry)
        return result

    def get_default_document_library(self, accountName, createSiteIfNotExists=False,
                                     siteCreationPriority=PersonalSiteCreationPriority.Low):
        """

        :param str accountName:
        :param bool createSiteIfNotExists:
        :param int siteCreationPriority:
        :return:
        """
        result = ClientResult(self.context)
        params = {"accountName": accountName,
                  "createSiteIfNotExists": createSiteIfNotExists,
                  "siteCreationPriority": siteCreationPriority}
        qry = ServiceOperationQuery(self, "GetDefaultDocumentLibrary", params, None, None, result)
        self.context.add_query(qry)
        return result

    def get_people_followed_by(self, account_name):
        """

        :type account_name: str
        :return: PersonPropertiesCollection
        """
        result = PersonPropertiesCollection(self.context)
        params = {"accountName": account_name}
        qry = ServiceOperationQuery(self, "GetPeopleFollowedBy", params, None, None, result)
        self.context.add_query(qry)
        return result

    def get_my_followers(self):
        """
        """
        return_type = PersonPropertiesCollection(self.context)
        qry = ServiceOperationQuery(self, "GetMyFollowers", None, None, None, return_type)
        self.context.add_query(qry)
        return return_type



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.comments.comment import Comment


class CommentCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(CommentCollection, self).__init__(context, Comment, resource_path)



from office365.entity import Entity
from office365.reports.internal.queries.create_report_query import create_report_query


class ReportRoot(Entity):
    """The resource that represents an instance of History Reports."""

    def get_email_activity_counts(self, period):
        """
        Enables you to understand the trends of email activity (like how many were sent, read, and received)
        in your organization.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """

        qry = create_report_query(self, "getEmailActivityCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_email_activity_user_counts(self, period):
        """
        Enables you to understand trends on the number of unique users who are performing email activities
        like send, read, and receive.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getEmailActivityUserCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_office365_activations_user_counts(self):
        """
        Get the count of Microsoft 365 activations on desktops and devices.

        """
        qry = create_report_query(self, "getOffice365ActivationsUserCounts")
        self.context.add_query(qry)
        return qry.return_type

    def get_onedrive_activity_file_counts(self, period):
        """
        Get the number of unique, licensed users that performed file interactions against any OneDrive account.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getOneDriveActivityFileCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_onedrive_activity_user_counts(self, period):
        """
        Get the trend in the number of active OneDrive users.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getOneDriveActivityUserCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_onedrive_activity_user_detail(self, period):
        """
        Get details about OneDrive activity by user.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getOneDriveActivityUserDetail", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_onedrive_usage_file_counts(self, period):
        """
        Get the total number of files across all sites and how many are active files. A file is considered active
        if it has been saved, synced, modified, or shared within the specified time period.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getOneDriveUsageFileCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_onedrive_usage_storage(self, period):
        """
        Get the trend on the amount of storage you are using in OneDrive for Business.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getOneDriveUsageStorage", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_sharepoint_activity_pages(self, period):
        """
        Get the number of unique pages visited by users.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getSharePointActivityPages", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_sharepoint_activity_user_counts(self, period):
        """
        Get the trend in the number of active users. A user is considered active if he or she has executed a
        file activity (save, sync, modify, or share) or visited a page within the specified time period.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getSharePointActivityUserCounts", period)
        self.context.add_query(qry)
        return qry.return_type

    def get_sharepoint_activity_user_detail(self, period):
        """
        Get details about SharePoint activity by user.

        :param str period: Specifies the length of time over which the report is aggregated.
            The supported values for {period_value} are: D7, D30, D90, and D180. These values follow the format
            Dn where n represents the number of days over which the report is aggregated. Required.
        """
        qry = create_report_query(self, "getSharePointActivityUserDetail", period)
        self.context.add_query(qry)
        return qry.return_type



import logging

LOGGING_SECRET_LVL = 5
LOGGING_SECRET_NAME = "SECDEBUG"


def ensure_debug_secrets():
    """Idempotent: add debug_secrets to Python's logging.Logger class."""
    if 'debug_secrets' not in logging.Logger.__dict__:
        def _log_secrets(self, *args, **kwargs):
            self.log(LOGGING_SECRET_LVL, *args, **kwargs)

        logging.Logger.debug_secrets = _log_secrets


# noinspection PyClassHasNoInit
class LoggerContext:
    """Superclass for all classes that require namespaced logger."""
    _logger = None

    @classmethod
    def logger(cls, method=None):
        if cls._logger is None:
            logger_name = cls.__module__ + '.' + cls.__name__
            cls._logger = logging.getLogger(logger_name)

        if method is not None:
            return cls._logger.getChild(method)
        return cls._logger



from office365.entity_collection import EntityCollection
from office365.onenote.entity_hierarchy_model import OnenoteEntityHierarchyModel
from office365.onenote.notebooks.notebook import Notebook
from office365.onenote.sections.section import OnenoteSection
from office365.runtime.resource_path import ResourcePath


class SectionGroup(OnenoteEntityHierarchyModel):
    """A section group in a OneNote notebook. Section groups can contain sections and section groups."""

    @property
    def section_groups_url(self):
        """The URL for the sectionGroups navigation property, which returns all the section groups in the section group.

        :rtype: str or None
        """
        return self.properties.get("sectionGroupsUrl", None)

    @property
    def sections_url(self):
        """The URL for the sections navigation property, which returns all the sections in the section group.

        :rtype: str or None
        """
        return self.properties.get("sectionsUrl", None)

    @property
    def parent_notebook(self):
        """The notebook that contains the section group. Read-only."""
        return self.get_property('parentNotebook',
                                 Notebook(self.context, ResourcePath("parentNotebook", self.resource_path)))

    @property
    def sections(self):
        """The sections in the section group. Read-only. Nullable.

        :rtype: EntityCollection
        """
        return self.get_property('sections',
                                 EntityCollection(self.context, OnenoteSection,
                                                  ResourcePath("sections", self.resource_path)))

    @property
    def section_groups(self):
        """Retrieve a list of onenoteSection objects from the specified notebook.

        :rtype: EntityCollection
        """
        return self.get_property('sectionGroups',
                                 EntityCollection(self.context, SectionGroup,
                                                  ResourcePath("sectionGroups", self.resource_path)))



from office365.runtime.client_value import ClientValue


class TimeSlot(ClientValue):
    """Represents a time slot for a meeting."""

    def __init__(self, start, end):
        """

        :param datetime.datetime start: The date, time, and time zone that a period begins.
        :param datetime.datetime end: The date, time, and time zone that a period ends.
        """
        super(TimeSlot, self).__init__()
        self.start = start
        self.end = end



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.queries.update_entity_query import UpdateEntityQuery
from office365.sharepoint.publishing.site_page_metadata import SitePageMetadata


class SitePage(SitePageMetadata):
    """Represents a site Page."""

    def checkout_page(self):
        """Checks out the current Site Page if it is available to be checked out."""
        site_page = SitePage(self.context)
        qry = ServiceOperationQuery(self, "CheckoutPage", None, None, None, site_page)
        self.context.add_query(qry)
        return site_page

    def copy(self):
        """Creates a copy of the current Site Page and returns the resulting new SitePage."""
        qry = ServiceOperationQuery(self, "Copy", None, None, None, None)
        self.context.add_query(qry)
        return self

    def discard_page(self):
        """Discards the current checked out version of the Site Page.  Returns the resulting SitePage after discard."""
        qry = ServiceOperationQuery(self, "DiscardPage", None, None, None, None)
        self.context.add_query(qry)
        return self

    def save_page(self, page_stream):
        """
        Updates the current Site Page with the provided pageStream content.

        :param str page_stream: The binary stream to save for the current Site Page.
        :return:
        """
        pass

    def save_page_as_draft(self, page_stream):
        """
        Updates the Site Page with the provided pageStream content and checks in a minor version if the page library
        has minor versions enabled.

        :param str page_stream: The binary stream to save for the current Site Page.
        :return:
        """
        pass

    def save_page_as_template(self):
        pass

    def publish(self):
        """
        Publishes a major version of the current Site Page.  Returns TRUE on success, FALSE otherwise.

        :return:
        """
        pass

    def share_page_preview_by_email(self, message, recipient_emails):
        pass



from office365.sharepoint.base_entity import BaseEntity


class Audit(BaseEntity):
    """
    Enables auditing of how site collections, sites, lists, folders, and list items are accessed, changed, and used.
    """
    pass



from office365.runtime.client_value import ClientValue


class SearchRequest(ClientValue):
    """A search request formatted in a JSON blob."""

    def __init__(self, query, entity_types=None):
        """
        :type query: office365.search.search_query.SearchQuery
        :type entity_types: list[str]
        """
        super(SearchRequest, self).__init__()
        self.query = query
        self.entityTypes = entity_types



from office365.onedrive.internal.queries.upload_session_query import UploadSessionQuery
from office365.onedrive.driveitems.driveItem import DriveItem
from office365.onedrive.driveitems.drive_item_uploadable_properties import DriveItemUploadableProperties
from office365.onedrive.internal.resourcepaths.resource_path_url import ResourcePathUrl


class ResumableFileUpload(UploadSessionQuery):
    """Create an upload session to allow your app to upload files up to the maximum file size. An upload session
    allows your app to upload ranges of the file in sequential API requests, which allows the transfer to be resumed
    if a connection is dropped while the upload is in progress. """

    def create_upload_session(self):
        item = DriveItemUploadableProperties()
        item.name = self.file_name
        return self.binding_type.create_upload_session(item)

    @property
    def binding_type(self):
        return DriveItem(self.context, ResourcePathUrl(self.file_name, self._binding_type.resource_path))

    @property
    def return_type(self):
        if self._return_type is None:
            self._return_type = DriveItem(self.context,
                                          ResourcePathUrl(self.file_name, self._binding_type.resource_path))
        return self._return_type



from office365.runtime.client_value import ClientValue


class DriveRecipient(ClientValue):
    """
    The DriveRecipient resource represents a person, group, or other recipient to
    share with using the invite action.
    """

    def __init__(self, alias=None, email=None, object_id=None):
        super(DriveRecipient, self).__init__()
        self.alias = alias
        self.email = email
        self.objectId = object_id

    @staticmethod
    def from_email(value):
        return DriveRecipient(email=value)



from office365.entity import Entity
from office365.runtime.queries.service_operation_query import ServiceOperationQuery


class IdentityApiConnector(Entity):
    """
    Represents API connectors in an Azure Active Directory (Azure AD) tenants.

    An API connector used in your Azure AD External Identities self-service sign-up user flows allows you to call
    an API during the execution of the user flow. An API connector provides the information needed to call an API
    including an endpoint URL and authentication. An API connector can be used at a specific step in a user flow
    to affect the execution of the user flow. For example, the API response can block a user from signing up,
    show an input validation error, or overwrite user collected attributes.
    """

    def upload_client_certificate(self, pkcs12Value, password):
        """Upload a PKCS 12 format key (.pfx) to an API connector's authentication configuration.
        The input is a base-64 encoded value of the PKCS 12 certificate contents.
        This method returns an apiConnector."""

        payload = {
            "pkcs12Value": pkcs12Value,
            "password": password
        }
        qry = ServiceOperationQuery(self, "uploadClientCertificate", None, payload, None, None)
        self.context.add_query(qry)
        return self



from office365.entity import Entity
from office365.onedrive.workbooks.worksheets.worksheet import WorkbookWorksheet
from office365.runtime.resource_path import ResourcePath


class WorkbookNamedItem(Entity):
    """Represents a defined name for a range of cells or value. Names can be primitive named objects
    (as seen in the type below), range object, reference to a range. This object can be used to obtain range
    object associated with names."""

    @property
    def name(self):
        """The name of the object. Read-only.

        :rtype str or None
        """
        return self.properties.get('name', None)

    @property
    def comment(self):
        """Represents the comment associated with this name.

        :rtype str or None
        """
        return self.properties.get('comment', None)

    @property
    def worksheet(self):
        """Returns the worksheet on which the named item is scoped to. Available only if the item is scoped
        to the worksheet. Read-only."""
        return self.properties.get('worksheet',
                                   WorkbookWorksheet(self.context, ResourcePath("worksheet", self.resource_path)))




from office365.entity import Entity


class ColumnLink(Entity):
    """A columnLink on a contentType attaches a site columnDefinition to that content type."""

    @property
    def name(self):
        """The name of the column in this content type.

        :rtype: str or None
        """
        return self.properties.get("name", None)



from office365.directory.directory_object import DirectoryObject
from office365.entity_collection import EntityCollection


class AppRoleAssignment(DirectoryObject):
    pass


class AppRoleAssignmentCollection(EntityCollection):

    def __init__(self, context, resource_path=None):
        super(AppRoleAssignmentCollection, self).__init__(context, AppRoleAssignment, resource_path)



from office365.runtime.client_value import ClientValue
from office365.sharepoint.search.queryResult import QueryResult


class SearchResult(ClientValue):

    def __init__(self):
        super(SearchResult, self).__init__()
        self.PrimaryQueryResult = QueryResult()
        self.ElapsedTime = None
        self.Properties = None
        self.SecondaryQueryResults = None
        self.SpellingSuggestion = None
        self.TriggeredRules = None

    @property
    def entity_type_name(self):
        return "Microsoft.Office.Server.Search.REST.SearchResult"



class ResourcePath(object):
    """OData resource path"""

    def __init__(self, segment, parent=None):
        """

        :type segment: str
        :type parent: ResourcePath or None
        """
        self._segment = segment
        self._parent = parent

    def to_url(self):
        """
        Builds url
        :rtype: str
        """
        current = self
        segments = []
        if self.require_closing_delimiter:
            segments.insert(0, current.delimiter)

        while current:
            segments.insert(0, current.segment)
            if current.parent:
                if current.delimiter_precedence == current.parent.delimiter_precedence:
                    segments.insert(0, "/")
                elif current.delimiter_precedence < current.parent.delimiter_precedence:
                    segments.insert(0, current.parent.delimiter)
                else:
                    segments.insert(0, current.delimiter)
            else:
                segments.insert(0, current.delimiter)
            current = current.parent
        return "".join(segments)

    @property
    def parent(self):
        return self._parent

    @property
    def segment(self):
        return self._segment

    @property
    def delimiter(self):
        return "/"

    @property
    def delimiter_precedence(self):
        return 1

    @property
    def require_closing_delimiter(self):
        return False



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.types.resource_path import ResourcePath as SPResPath


class MoveCopyUtil(BaseEntity):

    @staticmethod
    def copy_folder(context, srcUrl, destUrl, options):
        """

        :param office365.sharepoint.utilities.move_copy_options.MoveCopyOptions options:
        :param str srcUrl:
        :param str destUrl:
        :param office365.sharepoint.client_context.ClientContext context: client context
        """
        result = ClientResult(context)
        util = MoveCopyUtil(context)
        payload = {
            "srcUrl": srcUrl,
            "destUrl": destUrl,
            "options": options
        }
        qry = ServiceOperationQuery(util, "CopyFolder", None, payload, None, result)
        qry.static = True
        context.add_query(qry)
        return result

    @staticmethod
    def copy_folder_by_path(context, srcPath, destPath, options):
        """

        :param office365.sharepoint.utilities.move_copy_options.MoveCopyOptions options:
        :param str srcPath:
        :param str destPath:
        :param office365.sharepoint.client_context.ClientContext context: client context
        """
        result = ClientResult(context)
        util = MoveCopyUtil(context)
        payload = {
            "srcPath": SPResPath(srcPath),
            "destPath": SPResPath(destPath),
            "options": options
        }
        qry = ServiceOperationQuery(util, "CopyFolderByPath", None, payload, None, result)
        qry.static = True
        context.add_query(qry)
        return result

    @staticmethod
    def move_folder(context, srcUrl, destUrl, options):
        """

        :param office365.sharepoint.utilities.move_copy_options.MoveCopyOptions options:
        :param str srcUrl:
        :param str destUrl:
        :param office365.sharepoint.client_context.ClientContext context: client context
        """
        util = MoveCopyUtil(context)
        payload = {
            "srcUrl": srcUrl,
            "destUrl": destUrl,
            "options": options
        }
        qry = ServiceOperationQuery(util, "MoveFolder", None, payload, None, None)
        qry.static = True
        context.add_query(qry)
        return util

    @staticmethod
    def move_folder_by_path(context, srcPath, destPath, options):
        """

        :param office365.sharepoint.utilities.move_copy_options.MoveCopyOptions options:
        :param str srcPath:
        :param str destPath:
        :param office365.sharepoint.client_context.ClientContext context: client context
        """
        util = MoveCopyUtil(context)
        payload = {
            "srcPath": SPResPath(srcPath),
            "destPath": SPResPath(destPath),
            "options": options
        }
        qry = ServiceOperationQuery(util, "MoveFolderByPath", None, payload, None, None)
        qry.static = True
        context.add_query(qry)
        return util



from office365.directory.identities.identity_set import IdentitySet
from office365.onenote.entity_schema_object_model import OnenoteEntitySchemaObjectModel


class OnenoteEntityHierarchyModel(OnenoteEntitySchemaObjectModel):

    @property
    def display_name(self):
        """The name of the section."""
        return self.properties.get("displayName", None)

    @property
    def created_by(self):
        """Identity of the user, device, and application which created the item. Read-only."""
        return self.properties.get("createdBy", IdentitySet())

    @property
    def last_modified_by(self):
        """Identity of the user, device, and application which created the item. Read-only."""
        return self.properties.get("lastModifiedBy", IdentitySet())

    @property
    def last_modified_datetime(self):
        """Gets date and time the item was last modified."""
        return self.properties.get('lastModifiedDateTime', None)

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "createdBy": self.created_by,
                "lastModifiedBy": self.last_modified_by
            }
            default_value = property_mapping.get(name, None)
        return super(OnenoteEntityHierarchyModel, self).get_property(name, default_value)



from office365.sharepoint.base_entity import BaseEntity


class FeatureDefinition(BaseEntity):
    """Contains the base definition of a feature, including its name, ID, scope, and version."""
    pass



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.userprofiles.personProperties import PersonProperties


class PersonPropertiesCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(PersonPropertiesCollection, self).__init__(context, PersonProperties, resource_path)



from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.teams.channels.channel import Channel
from office365.teams.channels.channel_collection import ChannelCollection
from office365.teams.schedule.schedule import Schedule
from office365.teams.team_fun_settings import TeamFunSettings
from office365.teams.team_guest_settings import TeamGuestSettings
from office365.teams.team_member_settings import TeamMemberSettings
from office365.teams.team_messaging_settings import TeamMessagingSettings
from office365.teams.apps.teams_app_installation import TeamsAppInstallation
from office365.teams.operations.teams_async_operation import TeamsAsyncOperation
from office365.teams.teams_template import TeamsTemplate


class Team(Entity):
    """A team in Microsoft Teams is a collection of channel objects. A channel represents a topic, and therefore a
    logical isolation of discussion, within a team. """

    def delete_object(self):
        def _team_loaded():
            group = self.context.groups[self.id]
            group.delete_object(False)

        self.ensure_property("id", _team_loaded)
        return self

    @property
    def fun_settings(self):
        """Settings to configure use of Giphy, memes, and stickers in the team."""
        return self.properties.get('funSettings', TeamFunSettings())

    @property
    def member_settings(self):
        """Settings to configure whether members can perform certain actions, for example,
        create channels and add bots, in the team."""
        return self.properties.get('memberSettings', TeamMemberSettings())

    @property
    def guest_settings(self):
        """Settings to configure whether guests can create, update, or delete channels in the team."""
        return self.properties.get('guestSettings', TeamGuestSettings())

    @property
    def messaging_settings(self):
        """Settings to configure messaging and mentions in the team."""
        return self.properties.get('guestSettings', TeamMessagingSettings())

    @property
    def display_name(self):
        """The name of the team.

        :rtype: str or None
        """
        return self.properties.get('displayName', None)

    @property
    def description(self):
        """An optional description for the team.

        :rtype: str or None
        """
        return self.properties.get('description', None)

    @property
    def classification(self):
        """An optional label. Typically describes the data or business sensitivity of the team.
        Must match one of a pre-configured set in the tenant's directory.

        :rtype: str or None
        """
        return self.properties.get('classification', None)

    @property
    def is_archived(self):
        """Whether this team is in read-only mode."""
        return self.properties.get('isArchived', None)

    @property
    def visibility(self):
        """The visibility of the group and team. Defaults to Public."""
        return self.properties.get('visibility', None)

    @property
    def web_url(self):
        """A hyperlink that will go to the team in the Microsoft Teams client. This is the URL that you get when
        you right-click a team in the Microsoft Teams client and select Get link to team. This URL should be treated
        as an opaque blob, and not parsed."""
        return self.properties.get('webUrl', None)

    @property
    def created_datetime(self):
        """Timestamp at which the team was created."""
        return self.properties.get('createdDateTime', None)

    @property
    def channels(self):
        """The collection of channels & messages associated with the team.

        :rtype: ChannelCollection
        """
        return self.get_property('channels',
                                 ChannelCollection(self.context, ResourcePath("channels", self.resource_path)))

    @property
    def primary_channel(self):
        """
        The general channel for the team.

        :rtype: Channel
        """
        return self.get_property('primaryChannel',
                                 Channel(self.context, ResourcePath("primaryChannel", self.resource_path)))

    @property
    def schedule(self):
        """The schedule of shifts for this team.

        :rtype: Schedule
        """
        return self.get_property('schedule',
                                 Schedule(self.context, ResourcePath("schedule", self.resource_path)))

    @property
    def installed_apps(self):
        """The apps installed in this team.

        :rtype: EntityCollection
        """
        return self.get_property('installedApps',
                                 EntityCollection(self.context, TeamsAppInstallation,
                                                  ResourcePath("installedApps", self.resource_path)))

    @property
    def operations(self):
        """The async operations that ran or are running on this team.

        :rtype: EntityCollection
        """
        return self.get_property('operations',
                                 EntityCollection(self.context, TeamsAsyncOperation,
                                                  ResourcePath("installedApps", self.resource_path)))

    @property
    def template(self):
        """The template this team was created from

        :rtype: TeamsTemplate
        """
        return self.get_property('template',
                                 TeamsTemplate(self.context, ResourcePath("template", self.resource_path)))

    def archive(self):
        """Archive the specified team. When a team is archived, users can no longer send or like messages on any
        channel in the team, edit the team's name, description, or other settings, or in general make most changes to
        the team. Membership changes to the team continue to be allowed. """
        qry = ServiceOperationQuery(self, "archive")
        self.context.add_query(qry)
        return self

    def unarchive(self):
        """Restore an archived team. This restores users' ability to send messages and edit the team, abiding by
        tenant and team settings. """
        qry = ServiceOperationQuery(self, "unarchive")
        self.context.add_query(qry)
        return self

    def clone(self):
        """Create a copy of a team. This operation also creates a copy of the corresponding group. """
        qry = ServiceOperationQuery(self, "clone")
        self.context.add_query(qry)
        return self

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "installedApps": self.installed_apps,
                "primaryChannel": self.primary_channel
            }
            default_value = property_mapping.get(name, None)
        return super(Team, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(Team, self).set_property(name, value, persist_changes)
        # fallback: determine whether resource path is resolved
        if name == "id" and self._resource_path.segment == "team":
            self._resource_path = ResourcePath(value, ResourcePath("teams"))
        return self



from office365.runtime.client_value import ClientValue


class PublicationFacet(ClientValue):

    def __init__(self, level=None, versionId=None):
        """

        :param str level: The state of publication for this document. Either published or checkout. Read-only.
        :param str versionId: The unique identifier for the version that is visible to the current caller. Read-only.
        """
        super(PublicationFacet, self).__init__()
        self.level = level
        self.versionId = versionId



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity


class FileVersion(BaseEntity):
    """Represents a version of a File object."""

    def download(self, file_object):
        """Downloads the file version as a stream and save into a file.

        :type file_object: typing.IO
        """
        def _file_version_loaded():
            result = self.open_binary_stream()

            def _process_response(response):
                """
                :type response: requests.Response
                """
                response.raise_for_status()
                file_object.write(result.value)
            self.context.after_execute(_process_response)
        self.ensure_property("ID", _file_version_loaded)
        return self

    def open_binary_stream(self):
        """Opens the file as a stream."""
        return_stream = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "OpenBinaryStream", None, None, None, return_stream)
        self.context.add_query(qry)
        return return_stream

    def open_binary_stream_with_options(self, open_options):
        """Opens the file as a stream."""
        return_stream = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "OpenBinaryStreamWithOptions", [open_options], None, None, return_stream)
        self.context.add_query(qry)
        return return_stream

    @property
    def created_by(self):
        """Gets the user that created the file version."""
        from office365.sharepoint.principal.user import User
        return self.properties.get("CreatedBy", User(self.context, ResourcePath("CreatedBy", self.resource_path)))

    @property
    def id(self):
        """Gets a file version identifier"""
        return int(self.properties.get("ID", -1))

    @property
    def url(self):
        """Gets a value that specifies the relative URL of the file version based on the URL for the site or subsite."""
        return self.properties.get("Url", None)

    @property
    def version_label(self):
        """Gets a value that specifies the implementation specific identifier of the file."""
        return self.properties.get("VersionLabel", None)

    @property
    def is_current_version(self):
        """Gets a value that specifies whether the file version is the current version."""
        return self.properties.get("IsCurrentVersion", None)

    @property
    def checkin_comment(self):
        """Gets a value that specifies the check-in comment."""
        return self.properties.get("CheckInComment", None)

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "CreatedBy": self.created_by,
            }
            default_value = property_mapping.get(name, None)
        return super(FileVersion, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(FileVersion, self).set_property(name, value, persist_changes)
        if self._resource_path is None:
            if name == "ID":
                self._resource_path = ResourcePathServiceOperation(
                    "GetById", [value], self._parent_collection.resource_path)
        return self



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.contenttypes.content_type_id import ContentTypeId
from office365.sharepoint.contenttypes.field_link_collection import FieldLinkCollection
from office365.sharepoint.fields.field_collection import FieldCollection


class ContentType(BaseEntity):
    """Specifies a content type."""

    def reorder_fields(self, field_names):
        """
        The ReorderFields method is called to change the order in which fields appear in a content type.

        :param list[str] field_names:
        """
        qry = ServiceOperationQuery(self, "ReorderFields", field_names)
        self.context.add_query(qry)
        return self

    def update(self, update_children):
        """
        Updates the content type, and any child objects  of the content type if specified,
        with any changes made to the content type.

        :param bool update_children:
        """
        qry = ServiceOperationQuery(self, "Update", [update_children])
        self.context.add_query(qry)
        return self

    @property
    def id(self):
        """Specifies an identifier for the content type as specified in [MS-WSSTS] section 2.1.2.8.1.
        :rtype: ContentTypeId
        """
        return self.properties.get("Id", ContentTypeId())

    @property
    def string_id(self):
        """A string representation of the value of the Id
        :rtype: str
        """
        return self.properties.get("StringId", None)

    @property
    def name(self):
        """Gets the name of the content type.
        :rtype: str or None
        """
        return self.properties.get("Name", None)

    @name.setter
    def name(self, value):
        """Sets the name of the content type.
        :rtype: str or None
        """
        self.set_property("Name", value)

    @property
    def description(self):
        """Gets the description of the content type.
        :rtype: str or None
        """
        return self.properties.get("Description", None)

    @description.setter
    def description(self, value):
        """Sets the description of the content type.
        :rtype: str or None
        """
        self.set_property("Description", value)

    @property
    def group(self):
        """Gets the group of the content type.
        :rtype: str or None
        """
        return self.properties.get("Group", None)

    @group.setter
    def group(self, value):
        """Sets the group of the content type.
        :rtype: str or None
        """
        self.set_property("Group", value)

    @property
    def schema_xml(self):
        """Specifies the XML schema that represents the content type.
        :rtype: str or None
        """
        return self.properties.get("SchemaXml", None)

    @property
    def fields(self):
        """Gets a value that specifies the collection of fields for the content type."""
        return self.properties.get('Fields',
                                   FieldCollection(self.context, ResourcePath("Fields", self.resource_path)))

    @property
    def parent(self):
        """Gets the parent content type of the content type."""
        return self.properties.get('Parent',
                                   ContentType(self.context, ResourcePath("Parent", self.resource_path)))

    @property
    def field_links(self):
        """Specifies the collection of field links for the content type."""
        return self.properties.get('FieldLinks',
                                   FieldLinkCollection(self.context, ResourcePath("FieldLinks", self.resource_path)))

    def get_property(self, name, default_value=None):
        if name == "FieldLinks":
            default_value = self.field_links
        return super(ContentType, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(ContentType, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if name == "StringId" and self._resource_path is None:
            self._resource_path = ResourcePathServiceOperation(
                "getById", [value], self._parent_collection.resource_path)
        return self



from office365.entity import Entity
from office365.runtime.client_value_collection import ClientValueCollection


class ApplicationTemplate(Entity):
    """Represents an application in the Azure AD application gallery."""

    @property
    def display_name(self):
        """
        The name of the application.

        :rtype: str or None
        """
        return self.properties.get("displayName", None)

    @property
    def categories(self):
        """
        The list of categories for the application. Supported values can be: Collaboration, Business Management,
        Consumer, Content management, CRM, Data services, Developer services, E-commerce, Education, ERP, Finance,
        Health, Human resources, IT infrastructure, Mail, Management, Marketing, Media, Productivity,
        Project management, Telecommunications, Tools, Travel, and Web design & hosting.
        """
        return self.properties.get("categories", ClientValueCollection(str))



from office365.entity import Entity
from office365.onedrive.driveitems.publication_facet import PublicationFacet


class BaseItemVersion(Entity):

    @property
    def last_modified_datetime(self):
        """Gets date and time the item was last modified."""
        return self.properties.get('lastModifiedDateTime', 	None)

    @property
    def publication(self):
        """Indicates the publication status of this particular version. Read-only."""
        return self.properties.get('publication', PublicationFacet())



from office365.runtime.odata.v3.json_light_format import JsonLightFormat


class ClientValue(object):
    """Represent complex type.
    Complex types consist of a list of properties with no key, and can therefore only exist as properties of a
    containing entity or as a temporary value
    """

    def __init__(self):
        super(ClientValue, self).__init__()

    def set_property(self, k, v, persist_changes=True):
        prop_type = getattr(self, k, None)
        if isinstance(prop_type, ClientValue) and v is not None:
            if isinstance(v, list):
                [prop_type.set_property(i, p_v, persist_changes) for i, p_v in enumerate(v)]
            else:
                [prop_type.set_property(k, p_v, persist_changes) for k, p_v in v.items()]
            setattr(self, k, prop_type)
        else:
            setattr(self, k, v)

    def get_property(self, k):
        return getattr(self, k)

    def to_json(self, json_format=None):
        """
        :type json_format: office365.runtime.odata.odata_json_format.ODataJsonFormat or None
        """
        json = dict((k, v) for k, v in vars(self).items() if v is not None)
        for n, v in json.items():
            if isinstance(v, ClientValue):
                json[n] = v.to_json(json_format)

        if isinstance(json_format, JsonLightFormat) and json_format.is_verbose and self.entity_type_name is not None:
            json[json_format.metadata_type_tag_name] = {'type': self.entity_type_name}

        return json

    @property
    def entity_type_name(self):
        return type(self).__name__



from office365.runtime.client_value import ClientValue


class Image(ClientValue):
    """
    The Image resource groups image-related properties into a single structure. If a DriveItem has a
    non-null image facet, the item represents a bitmap image.

    Note: If the service is unable to determine the width and height of the image, the Image resource may be empty.
    """
    pass



from office365.directory.groups.group import Group
from office365.entity_collection import EntityCollection
from office365.runtime.queries.create_entity_query import CreateEntityQuery


class GroupCollection(EntityCollection):
    """Group's collection"""

    def __init__(self, context, resource_path=None):
        super(GroupCollection, self).__init__(context, Group, resource_path)

    def add(self, group_properties):
        """Create a Group resource.  You can create the following types of groups:
        Office 365 group (unified group)
        Security group

        :type group_properties: GroupProfile"""
        grp = Group(self.context)
        self.add_child(grp)
        qry = CreateEntityQuery(self, group_properties, grp)
        self.context.add_query(qry)
        return grp



from office365.runtime.client_value import ClientValue


class GroupSiteInfo(ClientValue):

    def __init__(self):
        super(GroupSiteInfo, self).__init__()
        self.SiteStatus = None
        self.SiteUrl = None
        self.DocumentsUrl = None
        self.ErrorMessage = None
        self.GroupId = None



from office365.runtime.auth.authentication_provider import AuthenticationProvider


class NetworkCredentialProvider(AuthenticationProvider):
    """Provides credentials for password-based authentication schemes such as basic authentication"""

    def __init__(self, username, password):
        super(NetworkCredentialProvider, self).__init__()
        self.userCredentials = (username, password)

    def authenticate_request(self, request):
        request.auth = self.userCredentials



from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity


class VideoServiceDiscoverer(BaseEntity):

    def __init__(self, context):
        super(VideoServiceDiscoverer, self).__init__(context, ResourcePath("SP.Publishing.VideoServiceDiscoverer"))




from office365.directory.extensions.extension import Extension
from office365.directory.profile_photo import ProfilePhoto
from office365.entity_collection import EntityCollection
from office365.outlook.calendar.email_address import EmailAddress
from office365.outlook.mail.item import Item
from office365.outlook.mail.physical_address import PhysicalAddress
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.resource_path import ResourcePath


class Contact(Item):
    """User's contact."""

    @property
    def manager(self):
        """
        The name of the contact's manager.
        :rtype: str or None
        """
        return self.properties.get("manager", None)

    @manager.setter
    def manager(self, value):
        """
        Sets name of the contact's manager.
        :type value: str
        """
        self.set_property("manager", value)

    @property
    def mobile_phone(self):
        """
        The contact's mobile phone number.
        :rtype: str or None
        """
        return self.properties.get("mobilePhone", None)

    @mobile_phone.setter
    def mobile_phone(self, value):
        """
        Sets contact's mobile phone number.
        :type value: str
        """
        self.set_property("mobilePhone", value)

    @property
    def home_address(self):
        return self.properties.get("homeAddress", PhysicalAddress())

    @property
    def email_addresses(self):
        """The contact's email addresses."""
        return self.properties.get("emailAddresses", ClientValueCollection(EmailAddress))

    @email_addresses.setter
    def email_addresses(self, value):
        """Sets contact's email addresses.

        :type value: list[str]
        """
        self.set_property("emailAddresses", value)

    @property
    def extensions(self):
        """The collection of open extensions defined for the contact. Nullable.

        :rtype: EntityCollection
        """
        return self.get_property('extensions',
                                 EntityCollection(self.context, Extension,
                                                  ResourcePath("extensions", self.resource_path)))

    @property
    def photo(self):
        """Optional contact picture. You can get or set a photo for a contact.

        :rtype: EntityCollection
        """
        return self.get_property('photo',
                                 ProfilePhoto(self.context, ResourcePath("photo", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "emailAddresses": self.email_addresses,
            }
            default_value = property_mapping.get(name, None)
        return super(Contact, self).get_property(name, default_value)

